# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import mock
import os
import six
import time

from oslo_utils import uuidutils

from searchlight.elasticsearch import ROLE_USER_FIELD
from searchlight.tests import functional
from searchlight.tests import utils as test_utils

TENANT1 = uuidutils.generate_uuid()
TENANT2 = uuidutils.generate_uuid()
TENANT3 = uuidutils.generate_uuid()

USER1 = uuidutils.generate_uuid()

fake_version_list = [test_utils.FakeVersion('2.1'),
                     test_utils.FakeVersion('2.1')]

nova_version_getter = 'novaclient.v2.client.versions.VersionManager.list'

MATCH_ALL = {"query": {"match_all": {}}, "sort": [{"name": {"order": "asc"}}]}
EMPTY_RESPONSE = {"hits": {"hits": [], "total": 0, "max_score": 0.0},
                  "_shards": {"successful": 0, "failed": 0, "total": 0},
                  "took": 1,
                  "timed_out": False}


class TestSearchApi(functional.FunctionalTest):
    """Test case for API functionality that's not plugin-specific, although
    it can use plugins for the sake of making requests
    """
    def _modify_policy_file(self, rules):
        with open(self.policy_file, 'r') as policy_file:
            existing_policy = json.load(policy_file)

        existing_policy.update(rules)

        with open(self.policy_file, 'w') as policy_file:
            json.dump(existing_policy, policy_file)

        time.sleep(2)

    def test_server_up(self):
        self.assertTrue(self.ping_server(self.api_port))

    def test_elasticsearch(self):
        """Index a document and check elasticsearch for it to check
        things are working.
        """
        images_plugin = self.initialized_plugins['OS::Glance::Image']
        doc_id = uuidutils.generate_uuid()
        doc = {
            "owner": TENANT1,
            "visibility": "public",
            "id": doc_id,
            "name": "owned by tenant 1",
            "owner": TENANT1,
            "created_at": "2016-04-06T12:48:18Z"
        }

        self._index(images_plugin,
                    [doc])

        doc["project_id"] = doc["owner"]
        # Test the raw elasticsearch response
        es_doc = self._get_elasticsearch_doc(
            images_plugin.alias_name_search,
            images_plugin.get_document_type(),
            doc_id)
        self.assertEqual(set(['admin', 'user']),
                         set(es_doc['_source'].pop(ROLE_USER_FIELD)))

        doc['members'] = []
        doc['image_type'] = 'image'
        self.assertEqual(doc, es_doc['_source'])

    def test_empty_results(self):
        """Test an empty dataset gets empty results."""
        response, json_content = self._search_request(MATCH_ALL, TENANT1)
        self.assertEqual(200, response.status)
        self.assertEqual([], self._get_hit_source(json_content))

    def test_nested_objects(self):
        """Test queries against documents with nested complex objects."""
        doc1 = {
            "owner": TENANT1,
            "visibility": "public",
            "objects": [],
            "tags": [],
            "namespace": "some.value1",
            "created_at": "2016-04-06T12:48:18Z",
            "properties": {"prop1": {"title": "hello"},
                           "prop2": {"title": "bye"}}
        }
        doc2 = {
            "owner": TENANT1,
            "visibility": "public",
            "namespace": "some.value2",
            "objects": [],
            "tags": [],
            "created_at": "2016-04-06T12:48:18Z",
            "properties": {"prop1": {"title": "something else"},
                           "prop2": {"title": "hello"}}
        }

        self._index(self.initialized_plugins['OS::Glance::Metadef'],
                    [doc1, doc2])

        # Convert the input docs into expected format
        additional_props = {
            'resource_types': [],
            'display_name': None,
            'description': None,
            'protected': None,
            'updated_at': None,
            'project_id': TENANT1
        }
        doc1.update(additional_props)
        doc2.update(additional_props)

        doc1["properties"] = [{"name": "prop1", "title": "hello"},
                              {"name": "prop2", "title": "bye"}]
        doc2["properties"] = [{"name": "prop1", "title": "something else"},
                              {"name": "prop2", "title": "hello"}]
        doc1["id"] = doc1["namespace"]
        doc1["name"] = doc1["namespace"]
        doc2["id"] = doc2["namespace"]
        doc2["name"] = doc2["namespace"]

        def get_nested(qs):
            return {
                "query": {
                    "nested": {
                        "path": "properties",
                        "query": {
                            "query_string": {"query": qs}
                        }
                    }
                },
                "sort": [{"namespace": {"order": "asc"}}]
            }

        # Expect this to match both documents
        querystring = "properties.name:prop1"
        query = get_nested(querystring)
        response, json_content = self._search_request(query,
                                                      TENANT1,
                                                      role="admin")
        self.assertEqual([doc1, doc2], self._get_hit_source(json_content))

        # Expect this to match only doc1
        querystring = "properties.name:prop1 AND properties.title:hello"
        query = get_nested(querystring)
        response, json_content = self._search_request(query,
                                                      TENANT1,
                                                      role="admin")
        self.assertEqual([doc1], self._get_hit_source(json_content))

        # Expect this not to match any documents, because it
        # doesn't properly match any nested objects
        querystring = "properties.name:prop1 AND properties.title:bye"
        query = get_nested(querystring)
        response, json_content = self._search_request(query,
                                                      TENANT1,
                                                      role="admin")
        self.assertEqual([], self._get_hit_source(json_content))

        # Expect a match with
        querystring = "properties.name:prop3 OR properties.title:bye"
        query = get_nested(querystring)
        response, json_content = self._search_request(query,
                                                      TENANT1,
                                                      role="admin")
        self.assertEqual([doc1], self._get_hit_source(json_content))

    def test_query_none(self):
        """Test search when query is not specified"""
        id_1 = uuidutils.generate_uuid()
        tenant1_doc = {
            "owner": TENANT1,
            "id": id_1,
            "visibility": "public",
            "name": "owned by tenant 1",
            "created_at": "2016-04-06T12:48:18Z"
        }

        self._index(self.initialized_plugins['OS::Glance::Image'],
                    [tenant1_doc])

        response, json_content = self._search_request({"all_projects": True},
                                                      TENANT1)
        self.assertEqual(200, response.status)
        tenant1_doc["members"] = []
        tenant1_doc["project_id"] = tenant1_doc["owner"]
        self.assertEqual([tenant1_doc], self._get_hit_source(json_content))

    def test_facet_exclude_fields(self):
        id_1 = uuidutils.generate_uuid()
        tenant1_doc = {
            "owner": TENANT1,
            "id": id_1,
            "visibility": "public",
            "name": "owned by tenant 1",
            "created_at": "2016-04-06T12:48:18Z"
        }

        self._index(self.initialized_plugins['OS::Glance::Image'],
                    [tenant1_doc])

        response, json_content = self._facet_request(
            TENANT1,
            include_fields=False,
            doc_type="OS::Nova::Server")
        self.assertEqual(200, response.status)
        self.assertEqual(set(["doc_count"]),
                         set(six.iterkeys(json_content["OS::Nova::Server"])))
        self.assertEqual(0, json_content["OS::Nova::Server"]["doc_count"])

        response, json_content = self._facet_request(
            TENANT1,
            include_fields=False,
            doc_type="OS::Glance::Image")
        self.assertEqual(1, json_content["OS::Glance::Image"]["doc_count"])

    def test_facet_rbac(self):
        tenant1_doc = {
            "owner": TENANT1,
            "id": uuidutils.generate_uuid(),
            "visibility": "private",
            "name": "owned by tenant 1",
            "container_format": "not-shown",
            "created_at": "2016-04-06T12:48:18Z"
        }
        tenant2_doc = {
            "owner": TENANT2,
            "id": uuidutils.generate_uuid(),
            "visibility": "private",
            "name": "owned by tenant 1",
            "container_format": "shown",
            "created_at": "2016-04-06T12:48:18Z"
        }
        with mock.patch('glanceclient.v2.image_members.Controller.list',
                        return_value=[]):
            self._index(self.initialized_plugins['OS::Glance::Image'],
                        [tenant1_doc, tenant2_doc])

        response, json_content = self._facet_request(
            TENANT1,
            include_fields=True,
            doc_type="OS::Glance::Image")
        self.assertEqual(1, json_content["OS::Glance::Image"]["doc_count"])
        facets = json_content["OS::Glance::Image"]["facets"]
        visibility_facet = list(filter(lambda f: f["name"] == "visibility",
                                       facets))[0]
        self.assertEqual([{"key": "private", "doc_count": 1}],
                         visibility_facet["options"])

    def test_facets(self):
        """Check facets for a non-nested field (status)"""
        servers_plugin = self.initialized_plugins['OS::Nova::Server']
        server1 = {
            u'addresses': {},
            u'flavor': {u'id': u'1'},
            u'id': u'6c41b4d1-f0fa-42d6-9d8d-e3b99695aa69',
            u'image': {u'id': u'a'},
            u'name': u'instance1',
            u'status': u'ACTIVE',
            u'tenant_id': TENANT1,
            u'created_at': u'2016-04-06T12:48:18Z',
            u'updated_at': u'2016-04-07T15:51:35Z',
            u'user_id': u'27f4d76b-be62-4e4e-aa33bb11cc55'
        }
        server2 = {
            u'addresses': {},
            u'flavor': {u'id': u'1'},
            u'id': u'08ca6c43-eea8-48d0-bbb2-30c50109d5d8',
            u'image': {u'id': u'a'},
            u'name': u'instance2',
            u'status': u'RESUMING',
            u'tenant_id': TENANT1,
            u'created_at': u'2016-04-06T12:48:18Z',
            u'updated_at': u'2016-04-07T15:51:35Z',
            u'user_id': u'27f4d76b-be62-4e4e-aa33bb11cc55'
        }
        server3 = {
            u'addresses': {},
            u'flavor': {u'id': u'1'},
            u'id': u'08ca6c43-f0fa-48d0-48d0-53453522cda4',
            u'image': {u'id': u'a'},
            u'name': u'instance1',
            u'status': u'ACTIVE',
            u'tenant_id': TENANT1,
            u'created_at': u'2016-04-06T12:48:18Z',
            u'updated_at': u'2016-04-07T15:51:35Z',
            u'user_id': u'27f4d76b-be62-4e4e-aa33bb11cc55'
        }
        with mock.patch(nova_version_getter, return_value=fake_version_list):
            self._index(
                servers_plugin,
                [test_utils.DictObj(**server1),
                 test_utils.DictObj(**server2),
                 test_utils.DictObj(**server3)])

        # Query facets with options.
        response, json_content = self._facet_request(
            TENANT1,
            doc_type="OS::Nova::Server")

        self.assertEqual(3, json_content['OS::Nova::Server']['doc_count'])

        expected = {
            u'name': u'status',
            u'options': [
                {u'doc_count': 2, u'key': u'active'},
                {u'doc_count': 1, u'key': u'resuming'},
            ],
            u'type': u'string'
        }

        status_facet = list(six.moves.filter(
            lambda f: f['name'] == 'status',
            json_content['OS::Nova::Server']['facets']
        ))[0]
        self.assertEqual(
            expected,
            status_facet,
        )

        # Query facets without options.
        response, json_content = self._facet_request(
            TENANT1,
            doc_type="OS::Nova::Server", exclude_options=True)

        self.assertEqual(3, json_content['OS::Nova::Server']['doc_count'])

        expected = {
            u'name': u'status',
            u'type': u'string'
        }

        status_facet = list(six.moves.filter(
            lambda f: f['name'] == 'status',
            json_content['OS::Nova::Server']['facets']
        ))[0]
        self.assertEqual(
            expected,
            status_facet,
        )

    def test_search_rbac(self):
        servers_plugin = self.initialized_plugins['OS::Nova::Server']
        server1 = {
            u'addresses': {},
            u'flavor': {u'id': u'1'},
            u'id': u'6c41b4d1-f0fa-42d6-9d8d-e3b99695aa69',
            u'image': {u'id': u'a'},
            u'name': u'instance1',
            u'status': u'ACTIVE',
            u'tenant_id': TENANT1,
            u'created_at': u'2016-04-06T12:48:18Z',
            u'updated_at': u'2016-04-07T15:51:35Z',
            u'user_id': u'27f4d76b-be62-4e4e-aa33bb11cc55'
        }
        server2 = {
            u'addresses': {},
            u'flavor': {u'id': u'1'},
            u'id': u'08ca6c43-eea8-48d0-bbb2-30c50109d5d8',
            u'image': {u'id': u'a'},
            u'name': u'instance2',
            u'status': u'RESUMING',
            u'tenant_id': TENANT2,
            u'created_at': u'2016-04-06T12:48:18Z',
            u'updated_at': u'2016-04-07T15:51:35Z',
            u'user_id': u'27f4d76b-be62-4e4e-aa33bb11cc55'
        }

        with mock.patch(nova_version_getter, return_value=fake_version_list):
            self._index(
                servers_plugin,
                [test_utils.DictObj(**server1),
                 test_utils.DictObj(**server2)])

        response, json_content = self._search_request(
            {"query": {"match_all": {}}, "type": "OS::Nova::Server"},
            TENANT1,
            role="member")
        self.assertEqual(1, json_content['hits']['total'])
        self.assertEqual(["6c41b4d1-f0fa-42d6-9d8d-e3b99695aa69"],
                         [h["_source"]["id"]
                          for h in json_content["hits"]["hits"]])

        response, json_content = self._search_request(
            {"query": {"match_all": {}}, "type": "OS::Nova::Server",
             "all_projects": True},
            TENANT1,
            role="admin")
        self.assertEqual(2, json_content['hits']['total'])

    def test_nested_facets(self):
        """Check facets for a nested field (networks.OS-EXT-IPS:type). We
        expect a single count per server matched, not per object in the
        'networks' field. Also check that for fields that are typed as
        'object' (not 'nested') they're marked appropriately
        """
        servers_plugin = self.initialized_plugins['OS::Nova::Server']
        server1 = {
            u'addresses': {
                u'net4': [
                    {u'addr': u'127.0.0.1',
                     u'OS-EXT-IPS:type': u'fixed',
                     u'version': 4},
                    {u'addr': u'127.0.0.1',
                     u'OS-EXT-IPS:type': u'fixed',
                     u'version': 4}
                ]
            },
            u'flavor': {u'id': u'1'},
            u'id': u'6c41b4d1-f0fa-42d6-9d8d-e3b99695aa69',
            u'image': {u'id': u'a'},
            u'name': u'instance1',
            u'created_at': u'2016-04-07T15:49:35Z',
            u'updated_at': u'2016-04-07T15:51:35Z',
            u'status': u'ACTIVE',
            u'tenant_id': TENANT1,
            u'user_id': u'27f4d76b-be62-4e4e-aa33bb11cc55'
        }

        server2 = {
            u'addresses': {
                u'net4': [
                    {u'addr': u'127.0.0.1',
                     u'OS-EXT-IPS:type': u'fixed',
                     u'version': 4},
                    {u'addr': u'127.0.0.1',
                     u'OS-EXT-IPS:type': u'floating',
                     u'version': 4}
                ]
            },
            u'flavor': {u'id': u'1'},
            u'id': u'08ca6c43-eea8-48d0-bbb2-30c50109d5d8',
            u'created_at': u'2016-04-07T15:49:35Z',
            u'updated_at': u'2016-04-07T15:51:35Z',
            u'image': {u'id': u'a'},
            u'name': u'instance2',
            u'status': u'ACTIVE',
            u'tenant_id': TENANT1,
            u'user_id': u'27f4d76b-be62-4e4e-aa33bb11cc55'
        }

        with mock.patch(nova_version_getter, return_value=fake_version_list):
            self._index(
                servers_plugin,
                [test_utils.DictObj(**server1),
                 test_utils.DictObj(**server2)])

        response, json_content = self._facet_request(
            TENANT1,
            doc_type="OS::Nova::Server")

        self.assertEqual(2, json_content['OS::Nova::Server']['doc_count'])

        self.assertEqual(['OS::Nova::Server'],
                         list(six.iterkeys(json_content)))

        # server1 has two fixed addresses (which should be rolled up into one
        # match). server2 has fixed and floating addresses.
        expected = {
            u'name': u'networks.OS-EXT-IPS:type',
            u'options': [
                {u'doc_count': 2, u'key': u'fixed'},
                {u'doc_count': 1, u'key': u'floating'},
            ],
            u'type': u'string',
            u'nested': True
        }
        fixed_network_facet = list(six.moves.filter(
            lambda f: f['name'] == 'networks.OS-EXT-IPS:type',
            json_content['OS::Nova::Server']['facets']
        ))[0]
        self.assertEqual(
            expected,
            fixed_network_facet,
        )

        # Check that 'image.id' (not nested but 'object') has nested=False
        expected = {
            u'name': u'image.id',
            u'type': u'string',
            u'nested': False,
            u'resource_type': u'OS::Glance::Image',
            u'options': [
                {u'doc_count': 2, u'key': u'a'}
            ]
        }
        image_facet = list(six.moves.filter(
            lambda f: f['name'] == 'image.id',
            json_content['OS::Nova::Server']['facets']
        ))[0]
        self.assertEqual(
            expected,
            image_facet,
        )

    def test_server_role_field_rbac(self):
        """Check that admins and users get different versions of documents"""
        doc_id = u'abc'
        s1 = {
            u'addresses': {},
            u'OS-DCF:diskConfig': u'MANUAL',
            u'OS-EXT-AZ:availability_zone': u'nova',
            u'OS-EXT-SRV-ATTR:host': u'devstack',
            u'OS-EXT-SRV-ATTR:hypervisor_hostname': u'devstack',
            u'OS-EXT-SRV-ATTR:instance_name': u'instance-00000001',
            u'id': doc_id,
            u'image': {u'id': u'a'},
            u'flavor': {u'id': u'1'},
            u'name': 'instance1',
            u'status': u'ACTIVE',
            u'tenant_id': TENANT1,
            u'user_id': USER1,
            u'created_at': u'2016-04-07T15:49:35Z',
            u'updated_at': u'2016-04-07T15:51:35Z',
        }

        servers_plugin = self.initialized_plugins['OS::Nova::Server']
        with mock.patch(nova_version_getter, return_value=fake_version_list):
            self._index(
                servers_plugin,
                [test_utils.DictObj(**s1)])

        response, json_content = self._search_request(MATCH_ALL,
                                                      TENANT1,
                                                      role="admin")
        self.assertEqual(200, response.status)
        self.assertEqual(1, len(json_content['hits']['hits']))
        hit = json_content['hits']['hits'][0]
        self.assertEqual(doc_id + "_ADMIN", hit['_id'])
        for k in ('OS-EXT-SRV-ATTR:host',
                  'OS-EXT-SRV-ATTR:hypervisor_hostname',
                  'OS-EXT-SRV-ATTR:instance_name'):
            self.assertIn(k, hit['_source'])

        # Now as a non admin
        response, json_content = self._search_request(MATCH_ALL,
                                                      TENANT1,
                                                      role="member")
        self.assertEqual(200, response.status)
        self.assertEqual(1, len(json_content['hits']['hits']))
        hit = json_content['hits']['hits'][0]
        self.assertEqual(doc_id + "_USER", hit['_id'])
        for k, v in hit.items():
            self.assertFalse(k.startswith('OS-EXT-SRV-ATTR:'),
                             'No protected attributes should be present')

        for field in (u'status', u'OS-DCF:diskConfig'):
            self.assertIn(field, hit['_source'])

    def test_role_fishing(self):
        """Run some searches to ward against 'fishing' type attacks such that
        'admin only' fields can't be searched by ordinary users
        """
        admin_field, admin_value = (u'OS-EXT-SRV-ATTR:host', u'devstack')

        doc_id = u'abc'
        s1 = {
            u'addresses': {},
            u'id': doc_id,
            u'image': {u'id': u'a'},
            u'flavor': {u'id': u'1'},
            u'name': 'instance1',
            u'status': u'ACTIVE',
            u'tenant_id': TENANT1,
            u'user_id': USER1,
            admin_field: admin_value,
            u'created_at': u'2016-04-07T15:49:35Z',
            u'updated_at': u'2016-04-07T15:51:35Z',
        }

        servers_plugin = self.initialized_plugins['OS::Nova::Server']
        with mock.patch(nova_version_getter, return_value=fake_version_list):
            self._index(
                servers_plugin,
                [test_utils.DictObj(**s1)])

        # For each of these queries (which are really looking for the same
        # thing) we expect a result for an admin, and no result for a user
        term_query = {'term': {admin_field: admin_value}}
        query_string = {'query_string': {'query': admin_value}}  # search 'all'
        query_string_field = {'query_string': {
            'default_field': admin_field, 'query': admin_value}}

        for query in (term_query, query_string, query_string_field):
            full_query = {'query': query}
            response, json_content = self._search_request(full_query,
                                                          TENANT1,
                                                          role="admin")
            self.assertEqual(200, response.status)
            self.assertEqual(1, json_content['hits']['total'],
                             "No results for: %s" % query)
            self.assertEqual(doc_id + '_ADMIN',
                             json_content['hits']['hits'][0]['_id'])

            # The same search should not work for users
            response, json_content = self._search_request(full_query,
                                                          TENANT1,
                                                          role="user")
            self.assertEqual(200, response.status)
            self.assertEqual(0, json_content['hits']['total'])

        # Run the same queries against 'name'; should get results
        term_query['term'] = {'name': 'instance1'}
        query_string['query_string']['query'] = 'instance1'
        query_string_field['query_string'] = {
            'default_field': 'name', 'query': 'instance1'
        }

        for query in (term_query, query_string, query_string_field):
            full_query = {'query': query}
            response, json_content = self._search_request(full_query,
                                                          TENANT1,
                                                          role="user")
            self.assertEqual(200, response.status)
            self.assertEqual(1, json_content['hits']['total'],
                             "No results for: %s %s" % (query, json_content))
            self.assertEqual(doc_id + '_USER',
                             json_content['hits']['hits'][0]['_id'])

    def test_resource_policy(self):
        servers_plugin = self.initialized_plugins['OS::Nova::Server']
        images_plugin = self.initialized_plugins['OS::Glance::Image']
        server_doc = {
            u'addresses': {},
            u'id': 'abcdef',
            u'name': 'instance1',
            u'status': u'ACTIVE',
            u'tenant_id': TENANT1,
            u'user_id': USER1,
            u'image': {u'id': u'a'},
            u'flavor': {u'id': u'1'},
            u'created_at': u'2016-04-07T15:49:35Z',
            u'updated_at': u'2016-04-07T15:51:35Z'
        }

        image_doc = {
            "owner": TENANT1,
            "id": "1234567890",
            "visibility": "public",
            "name": "image",
            "created_at": "2016-04-06T12:48:18Z"
        }

        with mock.patch(nova_version_getter, return_value=fake_version_list):
            self._index(servers_plugin, [test_utils.DictObj(**server_doc)])
        self._index(images_plugin, [image_doc])

        # Modify the policy file to restrict Nova servers
        self._modify_policy_file({"resource:OS::Nova::Server": "role:admin"})

        response, json_content = self._search_request(MATCH_ALL,
                                                      TENANT1,
                                                      role="user")
        self.assertEqual(1, json_content['hits']['total'])
        self.assertEqual('OS::Glance::Image',
                         json_content['hits']['hits'][0]['_type'])

        response, json_content = self._search_request(MATCH_ALL,
                                                      TENANT1,
                                                      role="admin")
        self.assertEqual(2, json_content['hits']['total'])
        self.assertEqual(set(['OS::Glance::Image', 'OS::Nova::Server']),
                         set([hit['_type']
                              for hit in json_content['hits']['hits']]))

        response, json_content = self._facet_request(TENANT1, role="user")
        self.assertNotIn('OS::Nova::Server', json_content)

        response, json_content = self._facet_request(TENANT1, role="admin")
        self.assertIn('OS::Nova::Server', json_content)
        self.assertIn('OS::Glance::Image', json_content)

        response, json_content = self._request('GET', '/search/plugins',
                                               TENANT1,
                                               role='user')
        self.assertEqual(
            0, len(list(filter(lambda p: p['type'] == 'OS::Nova::Server',
                               json_content['plugins']))))

        response, json_content = self._request('GET', '/search/plugins',
                                               TENANT1,
                                               role='admin')
        self.assertEqual(
            1, len(list(filter(lambda p: p['type'] == 'OS::Nova::Server',
                               json_content['plugins']))))

    def test_domain_token_401(self):
        tenant_id = None
        domain_id = "default"
        response, content = self._request(
            "POST", "/search", tenant_id, {"query": {"match_all": {}}},
            decode_json=False,
            extra_headers={"X-Domain-Id": domain_id})
        self.assertEqual(401, response.status)

    def test_search_version(self):
        id_1 = uuidutils.generate_uuid()
        tenant1_doc = {
            "owner": TENANT1,
            "id": id_1,
            "visibility": "public",
            "created_at": "2016-04-06T12:48:18Z",
            "updated_at": "2016-04-06T12:48:18Z"
        }
        self._index(self.initialized_plugins['OS::Glance::Image'],
                    [tenant1_doc])
        query = {"query": {"match_all": {}},
                 "type": "OS::Glance::Image",
                 "version": True}
        response, json_content = self._search_request(query,
                                                      TENANT1,
                                                      role="user")
        self.assertEqual(1, json_content['hits']['total'])
        self.assertIn('_version', json_content['hits']['hits'][0])

    def test_aggregations(self):
        servers_plugin = self.initialized_plugins['OS::Nova::Server']
        images_plugin = self.initialized_plugins['OS::Glance::Image']
        volumes_plugin = self.initialized_plugins['OS::Cinder::Volume']
        server_doc = {
            u'addresses': {},
            u'id': 'abcdef',
            u'name': 'instance1',
            u'status': u'ACTIVE',
            u'tenant_id': TENANT1,
            u'user_id': USER1,
            u'image': {u'id': u'a'},
            u'flavor': {u'id': u'1'},
            u'created_at': u'2016-04-07T15:49:35Z',
            u'updated_at': u'2016-04-07T15:51:35Z'
        }

        image_doc = {
            "owner": TENANT1,
            "id": "1234567890",
            "visibility": "public",
            "name": "for_instance1",
            "created_at": "2016-04-06T12:48:18Z"
        }

        volume_doc = {
            "os-vol-tenant-attr:tenant_id": TENANT1,
            "user_id": USER1,
            "id": "deadbeef",
            "name": "for_instance1",
            "created_at": "2016-04-06T12:48:18Z",
            "updated_at": "2016-04-06T12:48:18Z",
            "volume_type": "lvmdriver-1"
        }

        with mock.patch(nova_version_getter, return_value=fake_version_list):
            self._index(servers_plugin, [test_utils.DictObj(**server_doc)])
        self._index(images_plugin, [image_doc])
        self._index(volumes_plugin, [test_utils.DictObj(**volume_doc)])

        query = {
            "limit": 0,
            "query": {"match_all": {}},
            "aggregations": {
                "names": {"terms": {"field": "name"}},
                "earliest": {"min": {"field": "created_at"}}
            }
        }
        response, json_content = self._search_request(query,
                                                      TENANT1,
                                                      role="admin")
        expected_aggregations = {
            "names": {
                "doc_count_error_upper_bound": 0,
                "sum_other_doc_count": 0,
                "buckets": [
                    {"key": "for_instance1", "doc_count": 2},
                    {"key": "instance1", "doc_count": 1}
                ]
            },
            "earliest": {
                "value": 1459946898000.0,
                "value_as_string": "2016-04-06T12:48:18.000Z"}
        }
        # Expect a total count but no hit documents
        self.assertEqual({"total": 3, "max_score": 0.0, "hits": []},
                         json_content["hits"])
        # Should see "aggregations" at the top level
        self.assertIn("aggregations", json_content)
        self.assertEqual(expected_aggregations, json_content["aggregations"])

    def test_aggregation_rbac(self):
        servers_plugin = self.initialized_plugins['OS::Nova::Server']
        images_plugin = self.initialized_plugins['OS::Glance::Image']
        server_doc = {
            u'addresses': {},
            u'id': 'abcdef',
            u'name': 'instance1',
            u'status': u'ACTIVE',
            u'tenant_id': TENANT1,
            u'user_id': USER1,
            u'image': {u'id': u'a'},
            u'flavor': {u'id': u'1'},
            u'created_at': u'2016-04-07T15:49:35Z',
            u'updated_at': u'2016-04-07T15:51:35Z'
        }

        image_doc = {
            "owner": TENANT1,
            "id": "1234567890",
            "visibility": "public",
            "name": "for_instance1",
            "created_at": "2016-04-06T12:48:18Z"
        }

        with mock.patch(nova_version_getter, return_value=fake_version_list):
            self._index(servers_plugin, [test_utils.DictObj(**server_doc)])
        self._index(images_plugin, [image_doc])

        # Set type to ignore the image for all queries
        query = {
            "limit": 0,
            "type": "OS::Nova::Server",
            "query": {"match_all": {}},
            "aggregations": {
                "names": {"terms": {"field": "name"}}
            }
        }

        expected_aggregation = {
            "names": {
                "doc_count_error_upper_bound": 0,
                "sum_other_doc_count": 0,
                "buckets": [
                    {"key": "instance1", "doc_count": 1}
                ]
            }
        }

        expected_aggregation_no_results = {
            "names": {
                "doc_count_error_upper_bound": 0,
                "sum_other_doc_count": 0,
                "buckets": []
            }
        }

        # Expected aggregation results
        response, json_content = self._search_request(query,
                                                      TENANT1,
                                                      role="member")
        self.assertEqual(1, json_content['hits']['total'])
        self.assertEqual(expected_aggregation, json_content['aggregations'])

        # Now with a different tenant
        response, json_content = self._search_request(query,
                                                      TENANT2,
                                                      role="member")
        self.assertEqual(0, json_content['hits']['total'])
        self.assertEqual(expected_aggregation_no_results,
                         json_content['aggregations'])

        # Now with all_projects
        query['all_projects'] = True
        response, json_content = self._search_request(query,
                                                      TENANT2,
                                                      role="member")
        self.assertEqual(0, json_content['hits']['total'])
        self.assertEqual(expected_aggregation_no_results,
                         json_content['aggregations'])

        # Now admin all projects
        response, json_content = self._search_request(query,
                                                      TENANT2,
                                                      role="admin")
        self.assertEqual(1, json_content['hits']['total'])
        self.assertEqual(expected_aggregation, json_content['aggregations'])

    def test_global_aggregation_not_allowed(self):
        query = {
            "limit": 0,
            "query": {"match_all": {}},
            "aggregations": {
                "cheating": {
                    "global": {},
                    "aggs": {
                        "names": {"terms": {"field": "name"}}
                    }
                }
            }
        }

        response, json_content = self._search_request(query,
                                                      TENANT1,
                                                      role="admin",
                                                      decode_json=False)
        # TODO(sjmc7) Grr! https://bugs.launchpad.net/searchlight/+bug/1610398
        self.assertNotEqual(200, response.status)
        # self.assertEqual(403, response.status)

    def test_aggregation_policy(self):
        query = {
            "limit": 0,
            "query": {"match_all": {}},
            "aggregations": {
                "names": {"terms": {"field": "name"}}
            }
        }

        # Alter the default policy to allow aggregation only for admins
        self._modify_policy_file({'search:query:aggregations': 'role:admin'})

        response, json_content = self._search_request(query,
                                                      TENANT1,
                                                      role="member",
                                                      decode_json=False)
        self.assertEqual(403, response.status)

        response, json_content = self._search_request(query,
                                                      TENANT1,
                                                      role="admin")
        self.assertEqual(200, response.status)

        # Now let nobody use aggregations
        self._modify_policy_file({'search:query:aggregations': '!'})

        response, json_content = self._search_request(query,
                                                      TENANT1,
                                                      role="member",
                                                      decode_json=False)
        self.assertEqual(403, response.status)

        response, json_content = self._search_request(query,
                                                      TENANT1,
                                                      role="admin",
                                                      decode_json=False)
        self.assertEqual(403, response.status)

    def test_is_admin_project(self):
        """The X-IS-ADMIN-PROJECT header is the current solution to the
        problems caused by the overloaded 'admin' role. It's only set if
        configured in keystone, and defaults to True for back compatibility.
        """
        one_tenant_doc = {
            u'addresses': {},
            u'id': 'abcdef',
            u'tenant_id': TENANT1,
            u'user_id': USER1,
            u'status': 'ACTIVE',
            u'image': {u'id': u'a'},
            u'flavor': {u'id': u'1'},
            u'created_at': u'2016-04-07T15:49:35Z',
            u'updated_at': u'2016-04-07T15:51:35Z'
        }
        two_tenant_doc = {
            u'addresses': {},
            u'id': '12341234',
            u'tenant_id': TENANT2,
            u'user_id': USER1,
            u'status': 'ACTIVE',
            u'image': {u'id': u'a'},
            u'flavor': {u'id': u'1'},
            u'created_at': u'2016-04-07T15:49:35Z',
            u'updated_at': u'2016-04-07T15:51:35Z'
        }

        # A request with HTTP_X_IS_ADMIN_PROJECT=false should not allow a
        # TENANT1-scoped request to retrieve both servers if set to True it
        # should (assuming they also have the admin role)
        servers_plugin = self.initialized_plugins['OS::Nova::Server']
        with mock.patch(nova_version_getter, return_value=fake_version_list):
            self._index(servers_plugin, [test_utils.DictObj(**one_tenant_doc),
                                         test_utils.DictObj(**two_tenant_doc)])

        query = {"type": "OS::Nova::Server",
                 "all_projects": True,
                 "query": {"match_all": {}},
                 "sort": "id"}
        # First try admin request, is-admin-project false
        response, json_content = self._search_request(query,
                                                      TENANT1,
                                                      role="admin",
                                                      is_admin_project=False)
        self.assertEqual(['abcdef_USER'],
                         [h['_id'] for h in json_content['hits']['hits']])

        # Now try with is_admin_project True
        response, json_content = self._search_request(query,
                                                      TENANT1,
                                                      role="admin",
                                                      is_admin_project=True)
        self.assertEqual(['12341234_ADMIN', 'abcdef_ADMIN'],
                         [h['_id'] for h in json_content['hits']['hits']])

        # Now try with is_admin_project True but member role
        response, json_content = self._search_request(query,
                                                      TENANT1,
                                                      role="member",
                                                      is_admin_project=True)
        self.assertEqual(['abcdef_USER'],
                         [h['_id'] for h in json_content['hits']['hits']])


class TestServerServicePolicies(functional.FunctionalTest):
    def _write_policy_file(self, filename, rules):
        with open(os.path.join(self.conf_dir, filename), 'w') as pol_file:
            json.dump(rules, pol_file)

    def _additional_server_config(self):
        """Create some service policy files"""
        super(TestServerServicePolicies, self)._additional_server_config()
        self.api_server.service_policy_files = ','.join([
            'image:glance-policy.json',
            'compute:' + os.path.join(self.conf_dir, 'nova-policy.json'),
            'volume:' + os.path.join(self.conf_dir, 'cinder-policy.json')])

        self.api_server.service_policy_path = self.conf_dir

        self._write_policy_file(
            'glance-policy.json',
            {'is_admin': 'role:admin', 'get_images': 'rule:is_admin'})
        self._write_policy_file(
            'nova-policy.json',
            {'os_compute_api:servers:index': '',
             'os_compute_api:os-hypervisors': 'is_admin:True'})
        self._write_policy_file(
            'cinder-policy.json',
            {'admin_or_owner': 'role:admin or project_id:%(project_id)s',
             'volume:get_all': 'rule:admin_or_owner'})

    def test_server_starts(self):
        """Test that with policy changes, the server still starts"""
        self.assertTrue(self.ping_server(self.api_port))

    def test_service_policy_facet(self):
        response, json_content = self._facet_request(TENANT1,
                                                     role="user")
        self.assertEqual(200, response.status)
        self.assertIn('OS::Nova::Server', json_content)
        self.assertIn('OS::Cinder::Volume', json_content)
        self.assertNotIn('OS::Glance::Image', json_content)
        # Next one tests is_admin which is a magic context rule
        self.assertNotIn('OS::Nova::Hypervisor', json_content)

        response, json_content = self._facet_request(TENANT1,
                                                     role="admin")
        self.assertEqual(200, response.status)
        self.assertIn('OS::Nova::Server', json_content)
        self.assertIn('OS::Glance::Image', json_content)
        self.assertIn('OS::Cinder::Volume', json_content)
        # Magic context rule should pass this time
        self.assertIn('OS::Nova::Hypervisor', json_content)

    def test_service_policy_search(self):
        servers_plugin = self.initialized_plugins['OS::Nova::Server']
        images_plugin = self.initialized_plugins['OS::Glance::Image']
        volumes_plugin = self.initialized_plugins['OS::Cinder::Volume']
        server_doc = {
            u'addresses': {},
            u'id': 'abcdef',
            u'name': 'instance1',
            u'status': u'ACTIVE',
            u'tenant_id': TENANT1,
            u'user_id': USER1,
            u'image': {u'id': u'a'},
            u'flavor': {u'id': u'1'},
            u'created_at': u'2016-04-07T15:49:35Z',
            u'updated_at': u'2016-04-07T15:51:35Z'
        }

        image_doc = {
            "owner": TENANT1,
            "id": "1234567890",
            "visibility": "public",
            "name": "image",
            "created_at": "2016-04-06T12:48:18Z"
        }

        volume_doc = {
            "os-vol-tenant-attr:tenant_id": TENANT1,
            "user_id": USER1,
            "id": "deadbeef",
            "created_at": "2016-04-06T12:48:18Z",
            "updated_at": "2016-04-06T12:48:18Z",
            "volume_type": "lvmdriver-1"
        }

        with mock.patch(nova_version_getter, return_value=fake_version_list):
            self._index(servers_plugin, [test_utils.DictObj(**server_doc)])
        self._index(images_plugin, [image_doc])
        self._index(volumes_plugin, [test_utils.DictObj(**volume_doc)])

        response, json_content = self._search_request(MATCH_ALL,
                                                      TENANT1,
                                                      role="user")
        self.assertEqual(200, response.status)
        self.assertEqual(2, json_content['hits']['total'])
        self.assertEqual(
            set([server_doc['id'], volume_doc['id']]),
            set(h['id'] for h in self._get_hit_source(json_content)))

        response, json_content = self._search_request(MATCH_ALL,
                                                      TENANT1,
                                                      role="admin")
        self.assertEqual(200, response.status)
        self.assertEqual(3, json_content['hits']['total'])
        self.assertEqual(
            set([server_doc['id'], image_doc['id'], volume_doc['id']]),
            set(s['id'] for s in self._get_hit_source(json_content)))

    def test_service_policy_forbidden_none_allowed(self):
        query = {"type": "OS::Glance::Image", "query": {"match_all": {}}}
        response, json_content = self._search_request(query,
                                                      TENANT1,
                                                      role="user",
                                                      decode_json=False)
        self.assertEqual(403, response.status)



#!/usr/bin/env python

# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Routines for configuring Searchlight
"""

import logging
import os
import tempfile

from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_middleware import cors
from oslo_policy import policy
from paste import deploy

from searchlight.i18n import _
from searchlight.version import version_info as version


paste_deploy_opts = [
    cfg.StrOpt('flavor',
               help=_('Partial name of a pipeline in your paste configuration '
                      'file with the service name removed. For example, if '
                      'your paste section name is '
                      '[pipeline:searchlight-api-keystone] use the value '
                      '"keystone"')),
    cfg.StrOpt('api_paste_config', default="api-paste.ini",
               help=_('The API paste config file to use.')),
]

common_opts = [
    cfg.IntOpt('limit_param_default', default=25,
               help=_('Default value for the number of items returned by a '
                      'request if not specified explicitly in the request')),
    cfg.IntOpt('api_limit_max', default=1000,
               help=_('Maximum permissible number of items that could be '
                      'returned by a request')),
    cfg.HostAddressOpt('pydev_worker_debug_host',
                       help=_('The hostname/IP of the pydev '
                              'process listening for '
                              'debug connections')),
    cfg.IntOpt('pydev_worker_debug_port', default=5678,
               help=_('The port on which a pydev process is listening for '
                      'connections.')),
    cfg.StrOpt('metadata_encryption_key', secret=True,
               help=_('AES key for encrypting store \'location\' metadata. '
                      'This includes, if used, Swift or S3 credentials. '
                      'Should be set to a random string of length 16, 24 or '
                      '32 bytes')),
    cfg.StrOpt('digest_algorithm', default='sha1',
               help=_('Digest algorithm which will be used for digital '
                      'signature; the default is sha1 the default in Kilo '
                      'for a smooth upgrade process, and it will be updated '
                      'with sha256 in next release(L). Use the command '
                      '"openssl list-message-digest-algorithms" to get the '
                      'available algorithms supported by the version of '
                      'OpenSSL on the platform. Examples are "sha1", '
                      '"sha256", "sha512", etc.')),
]

CONF = cfg.CONF
CONF.register_opts(paste_deploy_opts, group='paste_deploy')
CONF.register_opts(common_opts)
policy.Enforcer(CONF)


def parse_args(args=None, usage=None, default_config_files=None):
    if "OSLO_LOCK_PATH" not in os.environ:
        lockutils.set_defaults(tempfile.gettempdir())

    CONF(args=args,
         project='searchlight',
         version=version.cached_version_string(),
         usage=usage,
         default_config_files=default_config_files)


def parse_cache_args(args=None):
    config_files = cfg.find_config_files(project='searchlight',
                                         prog='searchlight-cache')
    parse_args(args=args, default_config_files=config_files)


def _get_deployment_flavor(flavor=None):
    """
    Retrieve the paste_deploy.flavor config item, formatted appropriately
    for appending to the application name.

    :param flavor: if specified, use this setting rather than the
                   paste_deploy.flavor configuration setting
    """
    if not flavor:
        flavor = CONF.paste_deploy.flavor
    return '' if not flavor else ('-' + flavor)


def _get_deployment_config_file():
    """
    Retrieve the deployment_config_file config item, formatted as an
    absolute pathname.
    """
    config_path = cfg.CONF.find_file(
        cfg.CONF.paste_deploy['api_paste_config'])
    if config_path is None:
        return None

    return os.path.abspath(config_path)


def load_paste_app(app_name, flavor=None):
    """
    Builds and returns a WSGI app from a paste config file.

    We assume the last config file specified in the supplied ConfigOpts
    object is the paste config file, if conf_file is None.

    :param app_name: name of the application to load
    :param flavor: name of the variant of the application to load

    :raises RuntimeError when config file cannot be located or application
            cannot be loaded from config file
    """
    # append the deployment flavor to the application name,
    # in order to identify the appropriate paste pipeline
    app_name += _get_deployment_flavor(flavor)

    conf_file = _get_deployment_config_file()
    if conf_file is None:
        raise RuntimeError(_("Unable to locate config file"))

    try:
        logger = logging.getLogger(__name__)
        logger.debug("Loading %(app_name)s from %(conf_file)s",
                     {'conf_file': conf_file, 'app_name': app_name})

        app = deploy.loadapp("config:%s" % conf_file, name=app_name)

        # Log the options used when starting if we're in debug mode...
        if CONF.debug:
            CONF.log_opt_values(logger, logging.DEBUG)

        return app
    except (LookupError, ImportError) as e:
        msg = (_("Unable to load %(app_name)s from "
                 "configuration file %(conf_file)s."
                 "\nGot: %(e)r") % {'app_name': app_name,
                                    'conf_file': conf_file,
                                    'e': e})
        logger.error(msg)
        raise RuntimeError(msg)


def set_config_defaults():
    """This method updates all configuration default values."""
    set_cors_middleware_defaults()


def set_cors_middleware_defaults():
    """Update default configuration options for oslo.middleware."""
    # CORS Defaults
    # TODO(krotscheck): Update with https://review.openstack.org/#/c/285368/
    cfg.set_defaults(cors.CORS_OPTS,
                     allow_headers=['X-Auth-Token',
                                    'X-OpenStack-Request-ID'],
                     expose_headers=['X-OpenStack-Request-ID'],
                     allow_methods=['GET',
                                    'POST']
                     )



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import json
import logging
import novaclient.exceptions
import novaclient.v2.flavors
import six

from searchlight.elasticsearch.plugins import openstack_clients
from searchlight.elasticsearch.plugins import utils

LOG = logging.getLogger(__name__)


# All 'links' will also be removed
BLACKLISTED_FIELDS = set((u'progress', u'links'))
FLAVOR_ACCESS_FIELD = 'tenant_access'
FLAVOR_FIELDS_MAP = {
    'disabled': 'OS-FLV-DISABLED:disabled',
    'is_public': 'os-flavor-access:is_public',
    'ephemeral_gb': 'OS-FLV-EXT-DATA:ephemeral',
    'projects': 'tenant_access',
    'root_gb': 'disk',
    'memory_mb': 'ram'
}
FLAVOR_BLACKLISTED_FIELDS = ['vcpu_weight', 'flavorid']

EXTENDED_FIELDS = {'OS-EXT-STS:task_state': 'task_state',
                   'OS-EXT-STS:vm_state': 'state',
                   'OS-EXT-AZ:availability_zone': 'availability_zone',
                   'OS-EXT-SRV-ATTR:hypervisor_hostname': 'host_name',
                   'OS-EXT-SRV-ATTR:host': 'host'}


def _get_flavor_access(flavor):
    if flavor.is_public:
        return None
    try:
        n_client = openstack_clients.get_novaclient()
        return [access.tenant_id for access in
                n_client.flavor_access.list(flavor=flavor)] or None
    except novaclient.exceptions.Unauthorized:
        LOG.warning("Could not return tenant for %s; forbidden" %
                    flavor)
        return None


def serialize_nova_server(server):
    nc_client = openstack_clients.get_novaclient()
    if isinstance(server, six.text_type):
        server = nc_client.servers.get(server)

    LOG.debug("Serializing server %s for project %s",
              server.id, server.tenant_id)
    serialized = {k: v for k, v in server.to_dict().items()
                  if k not in BLACKLISTED_FIELDS}

    # Some enhancements
    serialized[u'owner'] = server.tenant_id
    serialized[u'project_id'] = server.tenant_id
    # Image is empty when the instance is booted from volume
    if isinstance(serialized[u'image'], dict):
        serialized[u'image'].pop(u'links', None)
    else:
        serialized.pop(u'image')
    serialized[u'flavor'].pop(u'links', None)

    sec_groups = serialized.pop(u'security_groups', [])
    serialized['security_groups'] = [s[u'name'] for s in sec_groups]

    _format_networks(server, serialized)

    utils.normalize_date_fields(serialized)

    serialized['status'] = serialized['status'].lower()

    # Pop the fault stracktrace if any - it's big
    fault = serialized.get('fault', None)
    if fault and isinstance(fault, dict):
        fault.pop('details', None)

    return serialized


# TODO(sjmc7) - if https://review.openstack.org/#/c/485525/ lands, remove this
# If it doesn't, make it more accurate
def _get_server_status(vm_state, task_state):
    # https://github.com/openstack/nova/blob/master/nova/api/openstack/common.py#L113
    # Simplified version of that
    if vm_state:
        vm_state = vm_state.lower()
    if task_state:
        task_state = task_state.lower()

    return {
        'active': 'active',
        'building': 'build',
        'stopped': 'shutoff',
        'resized': 'verify_resize',
        'paused': 'paused',
        'suspended': 'suspended',
        'rescued': 'rescue',
        'error': 'error',
        'deleted': 'deleted',
        'soft-delete': 'soft_deleted',
        'shelved': 'shelved',
        'shelved_offloaded': 'shelved_offloaded',
    }.get(vm_state)


def serialize_server_versioned(payload):
    # Based loosely on currently documented 1.1 InstanceActionPayload

    # Some transforms - maybe these could be made the same in nova?
    transform_keys = [('display_description', 'description'),
                      ('display_name', 'name'), ('uuid', 'id')]
    for src, dest in transform_keys:
        payload[dest] = payload.pop(src)

    copy_keys = [('tenant_id', 'project_id')]
    for src, dest in copy_keys:
        if src in payload:
            payload[dest] = payload.get(src)

    delete_keys = ['audit_period', 'node']
    for key in delete_keys:
        payload.pop(key, None)

    # We should denormalize this because it'd be better for searching
    flavor_id = payload.pop('flavor')['nova_object.data']['flavorid']
    payload['flavor'] = {'id': flavor_id}

    image_id = payload.pop('image_uuid')
    payload['image'] = {'id': image_id}

    # Translate the status, kind of. state and task_state will get
    # popped off shortly
    vm_state = payload.get('state', None)
    task_state = payload.get('task_state', None)
    payload['status'] = _get_server_status(vm_state, task_state)

    # Map backwards to the OS-EXT- attributes
    for ext_attr, simple_attr in EXTENDED_FIELDS.items():
        attribute = payload.pop(simple_attr, None)
        if attribute:
            payload[ext_attr] = attribute

    # Network information. This has to be transformed
    # TODO(sjmc7) Try to better reconcile this with the API format
    ip_addresses = [address['nova_object.data'] for address in
                    payload.pop("ip_addresses", [])]

    def map_address(addr):
        # TODO(sjmc7) Think this should be network name. Missing net type
        net = {
            "version": addr["version"],
            "name": addr["device_name"],
            "OS-EXT-IPS-MAC:mac_addr": addr["mac"],
        }
        if net["version"] == 4:
            net["ipv4_addr"] = addr["address"]
        else:
            net["ipv6_addr"] = addr["address"]
        return net

    payload["networks"] = [map_address(address) for address in ip_addresses]

    # Pop the fault stracktrace if any - it's big
    fault = payload.get('fault', None)
    if fault and isinstance(fault, dict):
        fault.pop('details', None)

    return payload


def serialize_nova_hypervisor(hypervisor, updated_at=None):
    serialized = hypervisor.to_dict()
    # The id for hypervisor is an integer, should be changed to
    # string.
    serialized['id'] = str(serialized['id'])
    # The 'cpu_info' field of hypervisor has changed from string
    # to JSON object in microversion 2.28, we should be able to
    # deal with JSON object here.
    if not isinstance(serialized['cpu_info'], dict):
        serialized['cpu_info'] = json.loads(serialized['cpu_info'])
    if not getattr(hypervisor, 'updated_at', None):
        serialized['updated_at'] = updated_at or utils.get_now_str()
    # TODO(lyj): Remove this once hypervisor notifications supported.
    for key in ['running_vms', 'vcpus_used', 'memory_mb_used', 'free_ram_mb',
                'free_disk_gb', 'local_gb_used', 'current_workload']:
        if key in serialized:
            serialized.pop(key)
    return serialized


def serialize_nova_flavor(flavor, updated_at=None):
    if hasattr(flavor, "to_dict"):
        serialized = {k: v for k, v in flavor.to_dict().items()
                      if k not in ("links")}
        serialized["extra_specs"] = flavor.get_keys()

        serialized[FLAVOR_ACCESS_FIELD] = _get_flavor_access(flavor)
    else:
        # This is a versioned flavor notification object
        serialized = copy.deepcopy(flavor)
        # Flavorid is a uuid like string.
        serialized['id'] = serialized['flavorid']
        # Extra specs and projects are added by update operation
        serialized['extra_specs'] = serialized.get('extra_specs') or {}
        serialized['projects'] = serialized.get('projects')

        # For consistent with the Flavor API response, we need to remove some
        # fields, and rename some fields key.
        for item in FLAVOR_BLACKLISTED_FIELDS:
            try:
                serialized.pop(item)
            except KeyError:
                pass

        for key, value in FLAVOR_FIELDS_MAP.items():
            try:
                serialized[value] = serialized.pop(key)
            except KeyError:
                pass

    if not serialized.get('updated_at'):
        serialized['updated_at'] = updated_at or utils.get_now_str()

    return serialized


def serialize_nova_servergroup(servergroup, updated_at=None):
    serialized = servergroup.to_dict()
    if not getattr(servergroup, 'updated_at', None):
        serialized['updated_at'] = updated_at or utils.get_now_str()
    return serialized


def _format_networks(server, serialized):
    networks = []

    # Keep the original as well
    addresses = copy.deepcopy(server.addresses)

    for net_name, ports in addresses.items():
        for port in ports:

            LOG.debug("Transforming net %s port %s for server %s",
                      net_name, port, server)
            addr = {u"name": net_name}
            port_address = port.pop(u'addr')
            if port[u'version'] == 4:
                port[u'ipv4_addr'] = port_address
            elif port[u'version'] == 6:
                port[u'ipv6_addr'] = port_address
            addr.update(port)
            networks.append(addr)
    serialized[u'networks'] = networks



# Copyright 2014 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import elasticsearch
import operator
from oslo_config import cfg

from searchlight.common import utils as common_utils

# Fields that require special handling for sort to avoid sorting
# on tokenized values
RAW_SORT_FIELDS = ('name',)

# It's not important exactly what this is, except that it's consistent
# across all plugins, and doesn't interfere with any real fields.
ROLE_USER_FIELD = "__searchlight-user-role"

search_opts = [
    cfg.ListOpt('hosts', default=['127.0.0.1:9200'],
                help='List of nodes where Elasticsearch instances are '
                     'running. A single node should be defined as an IP '
                     'address and port number.'),
    cfg.StrOpt('index_gc_deletes', default='300s',
               help='Time for which deleted documents are held in order to '
                    'prevent older, out-of-order updates causing them to be '
                    'created fresh in error.'),
    cfg.DictOpt('index_settings', default={},
                help='Dynamic index settings to be applied to new indices. '
                     'Format: opt1:val1,opt2:val2')
]

CONF = cfg.CONF
CONF.register_opts(search_opts, group='elasticsearch')


def get_api():
    es_hosts = CONF.elasticsearch.hosts
    es_api = elasticsearch.Elasticsearch(hosts=es_hosts)
    return es_api


class CatalogSearchRepo(object):

    def __init__(self, context, es_api):
        self.context = context
        self.es_api = es_api
        self.plugins = common_utils.get_search_plugins()
        self._plugins_list = self._get_plugin_list()

    def search(self, index, doc_type, query, from_,
               size, ignore_unavailable=True, **kwargs):
        return self.es_api.search(
            index=index,
            doc_type=doc_type,
            body=query,
            from_=from_,
            size=size,
            ignore_unavailable=ignore_unavailable,
            **kwargs)

    def plugins_info(self, doc_type):
        """Note: Even though we are using aliases to access ElasticSearch
        instead of indexes, we are still keeping the index field. This
        will allow the end-user to continue using other ElasticSearch
        tools which need an index to operate.
        """
        masked_plugins = filter(lambda p: p['type'] in doc_type,
                                self._plugins_list)
        return {
            'plugins': sorted(masked_plugins,
                              key=operator.itemgetter('type'))
        }

    def _get_plugin_list(self):
        plugin_list = []
        for plugin_type, plugin in self.plugins.items():
            plugin_list.append({
                'type': plugin.obj.get_document_type(),
                'alias-searching': plugin.obj.alias_name_search,
                'alias-indexing': plugin.obj.alias_name_listener
            })
        return plugin_list

    def facets(self, for_index, for_doc_types, all_projects, limit_terms,
               include_fields=True, exclude_options=False):
        """If include_fields is False, the only information returned will
        be a document count for the resource type.
        """
        facets = {}

        for resource_type, plugin in self.plugins.items():
            index_name = plugin.obj.alias_name_search
            doc_type = plugin.obj.get_document_type()

            if ((not for_index or index_name == for_index) and
                    (doc_type in for_doc_types)):

                # Add field facets if include_fields is true
                field_facets, doc_count = plugin.obj.get_facets(
                    self.context, all_projects, limit_terms,
                    include_fields=include_fields,
                    exclude_options=exclude_options)

                type_facets = {"doc_count": doc_count}
                if include_fields:
                    type_facets["facets"] = field_facets

                facets[resource_type] = type_facets

        return facets



# Copyright (c) 2016 Huawei Technology Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from searchlight.common import resource_types
from searchlight.elasticsearch.plugins import base
from searchlight.elasticsearch.plugins.nova \
    import notification_handler
from searchlight.elasticsearch.plugins.nova import serialize_nova_servergroup
from searchlight.elasticsearch.plugins import openstack_clients


class ServerGroupIndex(base.IndexBase):
    NotificationHandlerCls = notification_handler.ServerGroupHandler

    @classmethod
    def get_document_type(self):
        return resource_types.NOVA_SERVERGROUP

    def get_mapping(self):
        return {
            'dynamic': True,
            'properties': {
                'id': {'type': 'string', 'index': 'not_analyzed'},
                'project_id': {'type': 'string', 'index': 'not_analyzed'},
                'user_id': {'type': 'string', 'index': 'not_analyzed'},
                'name': {
                    'type': 'string',
                    'fields': {
                        'raw': {'type': 'string', 'index': 'not_analyzed'}
                    }
                },
                'policies': {'type': 'string', 'index': 'not_analyzed'},
                'members': {'type': 'string', 'index': 'not_analyzed'},
                'updated_at': {'type': 'date'},
                'metadata': {'type': 'object', 'properties': {}}
            },
            "_meta": {
                "project_id": {
                    "resource_type": resource_types.KEYSTONE_PROJECT
                },
                "user_id": {
                    "resource_type": resource_types.KEYSTONE_USER
                },
            }
        }

    @property
    def facets_with_options(self):
        return 'policies'

    @property
    def service_type(self):
        return 'compute'

    @property
    def resource_allowed_policy_target(self):
        return 'compute_extension:server_groups'

    def _get_rbac_field_filters(self, request_context):
        """Return any RBAC field filters to be injected into an indices
        query. Document type will be added to this list.
        """
        return [
            {'term': {'project_id': request_context.owner}},
        ]

    def get_objects(self):
        """Generator that lists all nova server groups."""
        for group in openstack_clients.get_novaclient(
        ).server_groups.list(all_projects=True):
            yield group

    def serialize(self, servergroup):
        return serialize_nova_servergroup(servergroup)



# Copyright 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import sys

import oslo_utils.strutils as strutils

from searchlight.i18n import _

try:
    import dns  # NOQA
except ImportError:
    dnspython_installed = False
else:
    dnspython_installed = True


def fix_greendns_ipv6():
    if dnspython_installed:
        # All of this is because if dnspython is present in your environment
        # then eventlet monkeypatches socket.getaddrinfo() with an
        # implementation which doesn't work for IPv6. What we're checking here
        # is that the magic environment variable was set when the import
        # happened.
        nogreendns = 'EVENTLET_NO_GREENDNS'
        flag = os.environ.get(nogreendns, '')
        if 'eventlet' in sys.modules and not strutils.bool_from_string(flag):
            msg = _("It appears that the eventlet module has been "
                    "imported prior to setting %s='yes'. It is currently "
                    "necessary to disable eventlet.greendns "
                    "if using ipv6 since eventlet.greendns currently "
                    "breaks with ipv6 addresses. Please ensure that "
                    "eventlet is not imported prior to this being set.")
            raise ImportError(msg % (nogreendns))

        os.environ[nogreendns] = 'yes'


fix_greendns_ipv6()



#    Copyright (c) 2016 Kylin Cloud
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import mock

from searchlight.listener import NotificationEndpoint
from searchlight.pipeline import PipelineManager
from searchlight.tests import functional
from searchlight.tests.functional import test_listener
from searchlight.tests import utils

TENANT1 = u"1816a16093df465dbc609cf638422a05"
TENANT_ID = u"1dd2c5280b4e45fc9d7d08a81228c891"

EV_TENANT = u"b1fbc598025841899f63333269abe1d2"

fake_version_list = [utils.FakeVersion('2.1'),
                     utils.FakeVersion('2.1')]

nova_server_getter = 'novaclient.v2.client.servers.ServerManager.get'
nova_version_getter = 'novaclient.v2.client.versions.VersionManager.list'


class TestNovaPlugins(functional.FunctionalTest):
    def setUp(self):
        super(TestNovaPlugins, self).setUp()
        self.hyper_plugin = self.initialized_plugins['OS::Nova::Hypervisor']
        self.hyper_objects = self._load_fixture_data('load/hypervisors.json')
        self.server_plugin = self.initialized_plugins['OS::Nova::Server']
        self.server_objects = self._load_fixture_data('load/servers.json')
        self.flavor_plugin = self.initialized_plugins['OS::Nova::Flavor']
        self.flavor_objects = self._load_fixture_data('load/flavors.json')
        self.server_groups_plugin = \
            self.initialized_plugins['OS::Nova::ServerGroup']
        self.server_groups_objects = \
            self._load_fixture_data('load/server_groups.json')

    @mock.patch(nova_version_getter, return_value=fake_version_list)
    def test_hypervisor_rbac(self, mock_version):
        self._index(self.hyper_plugin,
                    [utils.DictObj(**hyper) for hyper in self.hyper_objects])
        response, json_content = self._search_request(
            {"query": {"match_all": {}}, "all_projects": True},
            TENANT1, role="admin")
        self.assertEqual(200, response.status)
        self.assertEqual(2, json_content['hits']['total'])

        hits = json_content['hits']['hits']
        expected_sources = [{
            u'cpu_info': {u'arch': u'x86_64',
                          u'features': [u'pge', u'clflush'],
                          u'model': u'Nehalem',
                          u'topology': {u'cores': 20,
                                        u'sockets': 4,
                                        u'threads': 1},
                          u'vendor': u'Intel'},
            u'disk_available_least': 0,
            u'host_ip': u'192.168.1.11',
            u'hypervisor_hostname': u'host1',
            u'hypervisor_type': u'fake',
            u'hypervisor_version': 1000,
            u'id': u'1',
            u'local_gb': 600000,
            u'memory_mb': 800000,
            u'service': {u'disabled_reason': None,
                         u'host': u'host1',
                         u'id': 7},
            u'state': u'up',
            u'status': u'enabled',
            u'vcpus': 1000},
            {u'cpu_info': {u'arch': u'x86_64',
                           u'features': [u'pge', u'clflush'],
                           u'model': u'Nehalem',
                           u'topology': {u'cores': 10,
                                         u'sockets': 4,
                                         u'threads': 1},
                           u'vendor': u'Intel'},
             u'disk_available_least': 0,
             u'host_ip': u'192.168.1.12',
             u'hypervisor_hostname': u'host2',
             u'hypervisor_type': u'kvm',
             u'hypervisor_version': 1000,
             u'id': u'2',
             u'local_gb': 300000,
             u'memory_mb': 400000,
             u'service': {u'disabled_reason': None,
                          u'host': u'host2',
                          u'id': 8},
             u'state': u'up',
             u'status': u'enabled',
             u'vcpus': 500}]

        def process(hit):
            hit.pop('updated_at')
            return hit

        actual_sources = [process(hit['_source']) for hit in hits]
        self.assertEqual(expected_sources, actual_sources)

    @mock.patch(nova_version_getter, return_value=fake_version_list)
    def _index_data(self, mock_version):

        self._index(self.server_plugin,
                    [utils.DictObj(**server) for server in self.server_objects]
                    )

    def test_query(self):
        self._index_data()
        query = {
            "type": ["OS::Nova::Server"],
            "query": {
                "match_all": {}
            }
        }
        response, json_content = self._search_request(query,
                                                      TENANT_ID)
        self.assertEqual(200, response.status)
        self.assertEqual(1, json_content['hits']['total'])
        host_id = u'41d7069823d74c9ea8debda9a3a02bb00b2f7d53a0accd1f79429407'
        hits = json_content['hits']['hits']
        expected_sources = [{
            u'OS-DCF:diskConfig': u'MANUAL',
            u'OS-EXT-AZ:availability_zone': u'nova',
            u'OS-EXT-STS:power_state': 1,
            u'OS-EXT-STS:task_state': None,
            u'OS-EXT-STS:vm_state': u'active',
            u'OS-SRV-USG:launched_at': u'2016-03-08T08:40:22.000000',
            u'OS-SRV-USG:terminated_at': None,
            u'accessIPv4': u'',
            u'accessIPv6': u'',
            u'addresses': {u'private': [{
                           u'OS-EXT-IPS-MAC:mac_addr': u'fa:16:3e:55:7e:0d',
                           u'OS-EXT-IPS:type': u'fixed',
                           u'addr': u'10.0.0.2',
                           u'version': 4}]},
            u'config_drive': u'True',
            u'created': u'2016-03-08T08:39:33Z',
            u'created_at': u'2016-03-08T08:39:33Z',
            u'flavor': {u'id': u'1'},
            u'hostId': host_id,
            u'id': u'2adc3f33-8af9-43ef-8691-3d29f277e49b',
            u'image': {u'id': u'687c1900-a386-47c8-b591-b0083e7ca316'},
            u'key_name': u'testkey',
            u'metadata': {},
            u'name': u'test1',
            u'networks': [{u'OS-EXT-IPS-MAC:mac_addr': u'fa:16:3e:55:7e:0d',
                           u'OS-EXT-IPS:type': u'fixed',
                           u'ipv4_addr': u'10.0.0.2',
                           u'name': u'private',
                           u'version': 4}],
            u'os-extended-volumes:volumes_attached': [],
            u'owner': u'1dd2c5280b4e45fc9d7d08a81228c891',
            u'project_id': u'1dd2c5280b4e45fc9d7d08a81228c891',
            u'security_groups': [{u'name': u'default'}],
            u'status': u'active',
            u'tenant_id': u'1dd2c5280b4e45fc9d7d08a81228c891',
            u'updated': u'2016-03-08T08:40:22Z',
            u'user_id': u'7c97202cf58d43a9ab33016fc403f093'}]

        def process(hit):
            hit.pop('updated_at')
            return hit

        actual_sources = [process(hit['_source']) for hit in hits]
        self.assertEqual(expected_sources, actual_sources)

    def test_flavor_rbac(self):
        self._index(self.flavor_plugin,
                    [utils.FlavorDictObj(**flavor)
                     for flavor in self.flavor_objects]
                    )

        query = {
            "type": ["OS::Nova::Flavor"],
            "query": {
                "match_all": {}
            }
        }

        response, json_content = self._search_request(query, TENANT1)
        expected_sources = [{
            u'OS-FLV-DISABLED:disabled': False,
            u'OS-FLV-EXT-DATA:ephemeral': 0,
            u'disk': 1,
            u"tenant_id": u"",
            u'extra_specs': {},
            u'id': u'1',
            u'name': u'm1.tiny',
            u'os-flavor-access:is_public': True,
            u'ram': 512,
            u'rxtx_factor': 1.0,
            u'swap': u'',
            u'vcpus': 1}]
        hits = json_content['hits']['hits']
        for hit in hits:
            source = hit["_source"]
            source.pop("updated_at")
        actual_sources = [hit["_source"] for hit in hits]
        self.assertEqual(expected_sources, actual_sources)
        self.assertEqual(200, response.status)
        self.assertEqual(1, json_content['hits']['total'])

    def test_server_groups_rbac(self):
        self._index(self.server_groups_plugin,
                    [utils.DictObj(**server_group) for server_group in
                     self.server_groups_objects["server_groups"]])

        query = {
            "type": ["OS::Nova::ServerGroup"],
            "query": {
                "match_all": {}
            }
        }
        response, json_content = self._search_request(
            query,
            self.server_groups_objects["server_groups"][0]["project_id"]
        )
        expected_sources = [{
            u"user_id": u"944ff1aa607744ab9400acbf6be7f38a",
            u"policies": [u"affinity"],
            u"name": u"server_group_1",
            u"members": [u"8bf01fe7-1369-4059-92de-95ba11ff21dd"],
            u"project_id": u"d782f6257f0b484c97e9474b74db34a1",
            u"id": u"d0d017d6-0d41-4b32-b9f8-c43f039defc5",
            u"metadata": {}
        }]
        hits = json_content['hits']['hits']
        for hit in hits:
            source = hit["_source"]
            source.pop("updated_at")
        actual_sources = [hit["_source"] for hit in hits]
        self.assertEqual(expected_sources, actual_sources)
        self.assertEqual(200, response.status)
        self.assertEqual(1, json_content['hits']['total'])


class TestNovaListeners(test_listener.TestSearchListenerBase):
    def __init__(self, *args, **kwargs):
        super(TestNovaListeners, self).__init__(*args, **kwargs)
        self.server_events = self._load_fixture_data('events/servers.json')
        self.server_group_events = self._load_fixture_data(
            'events/server_group.json')
        self.flavor_events = self._load_fixture_data('events/flavor.json')

    def setUp(self):
        super(TestNovaListeners, self).setUp()
        self.servers_plugin = self.initialized_plugins['OS::Nova::Server']
        self.sg_plugin = self.initialized_plugins['OS::Nova::ServerGroup']
        self.flavor_plugin = self.initialized_plugins['OS::Nova::Flavor']

        notification_plugins = {
            plugin.document_type: utils.StevedoreMock(plugin)
            for plugin in (self.servers_plugin, self.sg_plugin,
                           self.flavor_plugin)}
        self.notification_endpoint = NotificationEndpoint(
            notification_plugins,
            PipelineManager(notification_plugins)
        )
        self.listener_alias = self.servers_plugin.alias_name_listener

    def test_server_group_create_delete(self):
        # Test #1: Create a server group.
        create_event = self.server_group_events["servergroup.create"]
        self._send_event_to_listener(create_event, self.listener_alias)

        query = {
            "query": {"match_all": {}},
            "type": "OS::Nova::ServerGroup"
        }

        response, json_content = self._search_request(query, EV_TENANT)

        self.assertEqual(200, response.status)
        self.assertEqual(1, json_content['hits']['total'])
        self.assertEqual(create_event['payload']['server_group_id'],
                         json_content['hits']['hits'][0]['_source']['id'])

        # Test #2: Add a new member to the server group.
        add_event = self.server_group_events['servergroup.addmember']
        self._send_event_to_listener(add_event, self.listener_alias)

        query = {"query": {"match_all": {}},
                 "type": "OS::Nova::ServerGroup"}
        response, json_content = self._search_request(query, EV_TENANT)

        self.assertEqual(200, response.status)
        # Verify the new member was added to the server group.
        members = (
            json_content['hits']['hits'][0]['_source']['members'])
        self.assertEqual(1, len(members))

        # Test #3: Delete the new member from the server group.
        delete_event = \
            self.server_group_events['compute.instance.delete.end']
        self._send_event_to_listener(delete_event, self.listener_alias)

        query = {"query": {"match_all": {}},
                 "type": "OS::Nova::ServerGroup"}
        response, json_content = self._search_request(query, EV_TENANT)

        self.assertEqual(200, response.status)
        # Verify the new member was deleted from the server group.
        members = (
            json_content['hits']['hits'][0]['_source']['members'])
        self.assertEqual(0, len(members))

        # Test #4: Delete the server group.
        delete_event = self.server_group_events['servergroup.delete']
        self._send_event_to_listener(delete_event, self.listener_alias)

        query = {"query": {"match_all": {}},
                 "type": "OS::Nova::ServerGroup"}
        response, json_content = self._search_request(query, EV_TENANT)

        self.assertEqual(200, response.status)
        self.assertEqual(0, json_content['hits']['total'])

    def test_flavor_create_update_delete(self):
        # Test #1: Create a flavor.
        create_event = self.flavor_events["flavor.create"]
        self._send_event_to_listener(create_event, self.listener_alias)

        query = {
            "query": {"match_all": {}},
            "type": "OS::Nova::Flavor"
        }

        response, json_content = self._search_request(query, EV_TENANT)

        self.assertEqual(200, response.status)
        self.assertEqual(1, json_content['hits']['total'])
        self.assertEqual(
            create_event['payload']['nova_object.data']['flavorid'],
            json_content['hits']['hits'][0]['_source']['id'])

        # Test #2: Update flavor with new extra specs.
        update_event = self.flavor_events['flavor.update']
        self._send_event_to_listener(update_event, self.listener_alias)

        query = {"query": {"match_all": {}},
                 "type": "OS::Nova::Flavor"}
        response, json_content = self._search_request(query, EV_TENANT)

        self.assertEqual(200, response.status)
        # Verify the new extra_specs was added to the flavor.
        extra_specs = (
            json_content['hits']['hits'][0]['_source']['extra_specs'])
        self.assertEqual({"key1": "value1"}, extra_specs)

        # Test #3: Delete the flavor.
        delete_event = self.flavor_events['flavor.delete']
        self._send_event_to_listener(delete_event, self.listener_alias)

        query = {"query": {"match_all": {}},
                 "type": "OS::Nova::Flavor"}
        response, json_content = self._search_request(query, EV_TENANT)

        self.assertEqual(200, response.status)
        self.assertEqual(0, json_content['hits']['total'])


class TestNovaUnversionedListener(test_listener.TestSearchListenerBase):
    def setUp(self):
        version_notifications = \
            'searchlight.elasticsearch.plugins.nova.notification_handler'\
            '.InstanceHandler._use_versioned_notifications'
        mock_versioned = mock.patch(version_notifications,
                                    return_value=False)
        mock_versioned.start()
        self.addCleanup(mock_versioned.stop)

        super(TestNovaUnversionedListener, self).setUp()
        self.servers_plugin = self.initialized_plugins['OS::Nova::Server']
        self.server_events = self._load_fixture_data('events/servers.json')

        sp = self.servers_plugin
        notification_plugins = {sp.document_type: utils.StevedoreMock(sp)}

        self.notification_endpoint = NotificationEndpoint(
            notification_plugins,
            PipelineManager(notification_plugins)
        )
        self.listener_alias = self.servers_plugin.alias_name_listener

    @mock.patch(nova_version_getter, return_value=fake_version_list)
    @mock.patch(nova_server_getter)
    def test_error_state_transition(self, mock_nova, mock_version_list):
        inst_id = "4b86f534-16db-4de8-8ce0-f1ee68894835"

        mock_nova.return_value = utils.DictObj(**{
            'id': inst_id,
            'name': 'test-error',
            'tenant_id': EV_TENANT,
            'addresses': {},
            'image': {'id': '1'},
            'flavor': {'id': 'a'},
            'status': 'ERROR',
            'created': '2016-08-31T23:32:11Z',
            'updated': '2016-08-31T23:32:11Z',
        })

        error_update = self.server_events[
            'instance-update-error-final'
        ]
        self._send_event_to_listener(error_update,
                                     self.listener_alias)
        result = self._verify_event_processing(error_update, owner=EV_TENANT)
        self._verify_result(error_update, ['tenant_id'], result)
        mock_nova.assert_called_with(inst_id)



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from searchlight.common import resource_types
from searchlight.elasticsearch.plugins import base
from searchlight.elasticsearch.plugins.nova \
    import notification_handler
from searchlight.elasticsearch.plugins.nova import serialize_nova_server
from searchlight.elasticsearch.plugins import openstack_clients


# TODO(sjmc7): Parameterize once we have plugin configs
LIST_LIMIT = 100


class ServerIndex(base.IndexBase):
    NotificationHandlerCls = notification_handler.InstanceHandler

    # Will be combined with 'admin_only_fields' from config
    # https://developer.openstack.org/api-ref/compute/?expanded=show-server-details-detail
    ADMIN_ONLY_FIELDS = ['OS-EXT-SRV-ATTR:*', 'host_status']

    @classmethod
    def get_document_type(self):
        return resource_types.NOVA_SERVER

    def get_mapping(self):
        return {
            'dynamic': True,
            'properties': {
                'id': {'type': 'string', 'index': 'not_analyzed'},
                'name': {
                    'type': 'string',
                    'fields': {
                        'raw': {'type': 'string', 'index': 'not_analyzed'}
                    }
                },
                'flavor': {
                    'type': 'object',
                    'properties': {
                        'id': {'type': 'string', 'index': 'not_analyzed'}
                    }
                },
                'owner': {'type': 'string', 'index': 'not_analyzed'},
                'tenant_id': {'type': 'string', 'index': 'not_analyzed'},
                'project_id': {'type': 'string', 'index': 'not_analyzed'},
                'user_id': {'type': 'string', 'index': 'not_analyzed'},
                'created': {'type': 'date'},
                'updated': {'type': 'date'},
                'created_at': {'type': 'date'},
                'updated_at': {'type': 'date'},
                'networks': {
                    'type': 'nested',
                    'properties': {
                        'name': {'type': 'string'},
                        'version': {'type': 'short'},
                        'OS-EXT-IPS-MAC:mac_addr': {
                            'type': 'string',
                            'index': 'not_analyzed'
                        },
                        'OS-EXT-IPS:type': {
                            'type': 'string',
                            'index': 'not_analyzed'
                        },
                        'ipv4_addr': {'type': 'ip'},
                        'ipv6_addr': {
                            'type': 'string',
                            'index': 'not_analyzed'
                        }
                    }
                },
                'image': {
                    'type': 'object',
                    'properties': {
                        'id': {'type': 'string', 'index': 'not_analyzed'}
                    }
                },
                'OS-EXT-AZ:availability_zone': {
                    'type': 'string',
                    'index': 'not_analyzed'
                },
                'OS-EXT-SRV-ATTR:hypervisor_hostname': {
                    'type': 'string',
                    'index': 'not_analyzed'
                },
                'OS-EXT-STS:vm_state': {
                    'type': 'string',
                    'index': 'not_analyzed'
                },
                'fault': {
                    'type': 'object',
                    'properties': {
                        'code': {'type': 'integer'},
                        'created': {'type': 'date'},
                        'message': {'type': 'string'},
                    }
                },
                # Nova gives security group names, where neutron ports
                # give ids in the same field. There's no solution that
                # maintains compatibility with both
                'security_groups': {'type': 'string', 'index': 'not_analyzed'},
                'status': {'type': 'string', 'index': 'not_analyzed'},
                # Nova adds/removes fields using microversion mechanism, check
                # http://git.openstack.org/cgit/openstack/nova/tree/nova/api/openstack/rest_api_version_history.rst
                # for detailed Nova microversion history.
                # Added in microversion 2.9
                'locked': {'type': 'string', 'index': 'not_analyzed'},
                # Added in microversion 2.16
                'host_status': {'type': 'string', 'index': 'not_analyzed'},
                # Added in microversion 2.19
                'description': {'type': 'string'},
                # Added in microversion 2.26
                'tags': {'type': 'string'},
            },
            "_meta": {
                "image.id": {
                    "resource_type": resource_types.GLANCE_IMAGE
                },
                "flavor.id": {
                    "resource_type": resource_types.NOVA_FLAVOR
                },
                "tenant_id": {
                    "resource_type": resource_types.KEYSTONE_PROJECT
                },
                "project_id": {
                    "resource_type": resource_types.KEYSTONE_PROJECT
                },
                "user_id": {
                    "resource_type": resource_types.KEYSTONE_USER
                },
                "OS-EXT-AZ:availability_zone": {
                    "resource_type": resource_types.NOVA_AVAILABILITY_ZONE
                },
                "security_groups": {
                    "resource_type": resource_types.NOVA_SECURITY_GROUP
                },
                "locked": {
                    "min_version": "2.9"
                },
                "host_status": {
                    "min_version": "2.16"
                },
                "description": {
                    "min_version": "2.19"
                },
                "tags": {
                    "min_version": "2.26"
                }
            },
        }

    @property
    def admin_only_fields(self):
        from_conf = super(ServerIndex, self).admin_only_fields
        return ServerIndex.ADMIN_ONLY_FIELDS + from_conf

    @property
    def facets_with_options(self):
        return ('OS-EXT-AZ:availability_zone',
                'status', 'image.id', 'flavor.id', 'networks.name',
                'networks.OS-EXT-IPS:type', 'networks.version',
                'security_groups', 'host_status', 'locked')

    @property
    def facets_excluded(self):
        """A map of {name: allow_admin} that indicate which
        fields should not be offered as facet options, or those that should
        only be available to administrators.
        """
        return {'tenant_id': True, 'project_id': True, 'host_status': True,
                'created': False, 'updated': False,
                'OS-EXT-SRV-ATTR:hypervisor_hostname': True, 'fault': False}

    @property
    def resource_allowed_policy_target(self):
        return 'os_compute_api:servers:index'

    @property
    def service_type(self):
        return 'compute'

    def _get_rbac_field_filters(self, request_context):
        """Return any RBAC field filters to be injected into an indices
        query. Document type will be added to this list.
        """
        return [
            {'term': {'tenant_id': request_context.owner}},
        ]

    def get_objects(self):
        """Generator that lists all nova servers owned by all tenants."""
        has_more = True
        marker = None
        while has_more:
            servers = openstack_clients.get_novaclient().servers.list(
                limit=LIST_LIMIT,
                search_opts={'all_tenants': True},
                marker=marker
            )

            if not servers:
                # Definitely no more; break straight away
                break

            # servers.list always returns a list so we can grab the last id
            has_more = len(servers) == LIST_LIMIT
            marker = servers[-1].id

            for server in servers:
                yield server

    def serialize(self, server):
        return serialize_nova_server(server)

    def filter_result(self, hit, request_context):
        super(ServerIndex, self).filter_result(hit, request_context)

        # Reverse the change we make to security groups in serialize() to
        # maintain compatibility with the nova API response
        source = hit['_source']
        security_groups = source.pop('security_groups', None)
        if security_groups is not None:
            source['security_groups'] = [{"name": sg}
                                         for sg in security_groups]



# Copyright (c) 2016 Hewlett-Packard Enterprise Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import mock
from oslo_utils import uuidutils

from searchlight.common import utils
from searchlight.elasticsearch.plugins.neutron import\
    ports as port_plugin
import searchlight.tests.unit.utils as unit_test_utils
import searchlight.tests.utils as test_utils


USER1 = u'27f4d76b-be62-4e4e-aa33bb11cc55'
ID1 = u'813dd936-663e-4e5b-877c-986021b73e2c'
TENANT1 = u'8eaac046b2c44ab99246cb0850c7f06d'
NETWORK1 = u'bc0adf22-3aef-4e7b-8b99-12670b5a76b5'
UUID_PORT_ID = uuidutils.generate_uuid()
_now_str = utils.isotime(datetime.datetime.utcnow())


def _create_port_fixture(port_id, tenant_id, network_id, **kwargs):
    port = {
        u'admin_state_up': True,
        u'allowed_address_pairs': [],
        u'binding:host_id': u'devstack',
        u'binding:profile': {},
        u'binding:vif_details': {u'ovs_hybrid_plug': True,
                                 u'port_filter': True},
        u'binding:vif_type': u'ovs',
        u'binding:vnic_type': u'normal',
        u'device_id': None,
        u'device_owner': 'compute:None',
        u'dns_assignment': [{
            u'fqdn': u'host-fd70-b64a-666f--1.openstacklocal.',
            u'hostname': u'host-fd70-b64a-666f--1',
            u'ip_address': u'fd70:b64a:666f::1'
        }],
        u'dns_name': u'',
        u'extra_dhcp_opts': [],
        u'fixed_ips': [{
            u'ip_address': u'fd70:b64a:666f::1',
            u'subnet_id': u'356bbedc-2bb2-467a-b380-8bfd0ed2fbdf'
        }],
        u'id': port_id,
        u'mac_address': u'fa:16:3e:19:c8:f5',
        u'name': u'',
        u'network_id': network_id,
        u'port_security_enabled': True,
        u'security_groups': [],
        u'status': u'ACTIVE',
        u'tenant_id': tenant_id,
        u'updated_at': _now_str,
        u'created_at': _now_str
    }
    port.update(**kwargs)
    return port


class TestPortLoaderPlugin(test_utils.BaseTestCase):
    def setUp(self):
        super(TestPortLoaderPlugin, self).setUp()
        self.plugin = port_plugin.PortIndex()
        self._create_fixtures()

    def _create_fixtures(self):
        self.port = _create_port_fixture(ID1, TENANT1, NETWORK1)
        self.none_port = _create_port_fixture(UUID_PORT_ID, TENANT1, NETWORK1,
                                              device_owner=None)
        self.dhcp_port = _create_port_fixture(UUID_PORT_ID, TENANT1, NETWORK1,
                                              device_owner='network:dhcp')
        # device_owner is 'compute:*', indexed.
        self.indexed_ports = [self.port]
        # device_owner is not 'compute:*', not indexed.
        self.ignored_ports = [self.dhcp_port, self.none_port]

    def test_default_index_name(self):
        self.assertEqual('searchlight', self.plugin.resource_group_name)

    def test_document_type(self):
        self.assertEqual('OS::Neutron::Port',
                         self.plugin.get_document_type())

    def test_rbac_filter_admin_role(self):
        fake_request = unit_test_utils.get_fake_request(
            USER1, TENANT1, '/v1/search', is_admin=True
        )
        rbac_terms = self.plugin._get_rbac_field_filters(fake_request.context)
        expected_rbac = [
            {'term': {'tenant_id': TENANT1}},
            {
                'has_parent': {
                    'type': self.plugin.parent_plugin_type(),
                    'query': {
                        "bool": {
                            "should": [
                                {'term': {'shared': True}},
                                {'term': {'router:external': True}}
                            ]
                        }
                    }
                }
            }
        ]
        self.assertEqual(expected_rbac, rbac_terms)

    def test_rbac_filter_non_admin_role(self):
        fake_request = unit_test_utils.get_fake_request(
            USER1, TENANT1, '/v1/search', is_admin=False
        )
        rbac_terms = self.plugin._get_rbac_field_filters(fake_request.context)
        self.assertEqual(
            [{"term": {"tenant_id": TENANT1}}],
            rbac_terms
        )

    def test_admin_only_fields(self):
        admin_only_fields = self.plugin.admin_only_fields
        self.assertEqual(['binding:*'], admin_only_fields)

    def test_indexed_ports(self):
        with mock.patch('neutronclient.v2_0.client.Client.list_ports',
                        return_value={'ports': self.indexed_ports}):
            listed_objects = list(self.plugin.get_objects())
            self.assertEqual([self.port], listed_objects)

    def test_ignored_ports(self):
        with mock.patch('neutronclient.v2_0.client.Client.list_ports',
                        return_value={'ports': self.ignored_ports}):
            listed_objects = list(self.plugin.get_objects())
            self.assertEqual([], listed_objects)

    def test_serialize(self):
        serialized = self.plugin.serialize(self.port)
        # project id should get copied from tenant_id
        self.assertEqual(TENANT1, serialized['project_id'])



# Copyright (c) 2016 Hewlett-Packard Enterprise Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import mock

from oslo_utils import uuidutils

from elasticsearch import helpers
from searchlight.elasticsearch.plugins.neutron import\
    security_groups as securitygroups_plugin
import searchlight.tests.unit.utils as unit_test_utils
import searchlight.tests.utils as test_utils

USER1 = uuidutils.generate_uuid()
ID1 = uuidutils.generate_uuid()
TENANT1 = uuidutils.generate_uuid()
_now_str = datetime.datetime.isoformat(datetime.datetime.utcnow())


def _secgroup_fixture(secgroup_id, tenant_id, name, **kwargs):
    fixture = {
        "security_group_rules": [{
            "remote_group_id": "855560aa-ff43-ee09-993e-6609342abccd",
            "direction": "ingress",
            "protocol": "ipv4",
            "description": "",
            "protocol": "",
            "ethertype": "",
            "remote_ip_prefix": "",
            "port_range_max": "",
            "port_range_min": "",
            "security_group_id": "99347aac-4412-b809-5112-434901123a9c",
            "tenant_id": tenant_id,
            "id": secgroup_id,
        }],
        "id": secgroup_id,
        "tenant_id": tenant_id,
        "name": name,
        "description": ""
    }
    fixture.update(**kwargs)
    return fixture


def _secgrouprule_fixture(secgroup_id, tenant_id, **kwargs):
    fixture = {
        "security_group_rule": {
            "remote_group_id": "855560aa-ff43-ee09-993e-6609342abccd",
            "direction": "ingress",
            "protocol": "ipv4",
            "description": "",
            "ethertype": "",
            "remote_ip_prefix": "",
            "port_range_max": "",
            "port_range_min": "",
            "security_group_id": secgroup_id,
            "tenant_id": tenant_id,
            "id": secgroup_id,
        }
    }
    fixture.update(**kwargs)
    return fixture


class TestSecurityGroupLoaderPlugin(test_utils.BaseTestCase):
    def setUp(self):
        super(TestSecurityGroupLoaderPlugin, self).setUp()
        self.plugin = securitygroups_plugin.SecurityGroupIndex()
        self._create_fixtures()

    def _create_fixtures(self):
        """Create a security group to test data serialization. """
        self.security_group = _secgroup_fixture(secgroup_id=ID1,
                                                tenant_id=TENANT1,
                                                name="test-secgroup")

    def test_document_type(self):
        self.assertEqual('OS::Neutron::SecurityGroup',
                         self.plugin.get_document_type())

    def test_rbac_filter(self):
        fake_request = unit_test_utils.get_fake_request(
            USER1, TENANT1, '/v1/search', is_admin=False
        )
        rbac_terms = self.plugin._get_rbac_field_filters(fake_request.context)
        self.assertEqual(
            [{"term": {"tenant_id": TENANT1}}],
            rbac_terms
        )

    def test_notification_events(self):
        handler = self.plugin.get_notification_handler()
        self.assertEqual(
            set(['security_group.create.end',
                 'security_group.delete.end',
                 'security_group_rule.create.end',
                 'security_group_rule.delete.end']),
            set(handler.get_event_handlers().keys())
        )

    def test_rule_update_exception(self):
        # Set up the return documents.
        payload = _secgrouprule_fixture(ID1, TENANT1)
        doc = {'_source': {'security_group_rules': [], 'id': 1},
               '_version': 1}

        handler = self.plugin.get_notification_handler()
        with mock.patch.object(self.plugin.index_helper,
                               'get_document') as mock_get:
            with mock.patch.object(self.plugin.index_helper,
                                   'save_document') as mock_save:
                mock_get.return_value = doc
                exc_obj = helpers.BulkIndexError(
                    "Version conflict", [{'index': {
                        "_id": "1", "error": "Some error", "status": 409}}]
                )

                # 1 retry (exception).
                mock_save.side_effect = [exc_obj, {}]
                handler.create_or_update_rule(
                    'security_group_rule.create.end', payload, None)
                # 1 retry +  1 success = 2 calls.
                self.assertEqual(2, mock_get.call_count)
                self.assertEqual(2, mock_save.call_count)

                # 24 retries (exceptions) that exceed the retry limit.
                # Not all retries will be used.
                mock_get.reset_mock()
                mock_save.reset_mock()
                mock_save.side_effect = [exc_obj, exc_obj, exc_obj, exc_obj,
                                         exc_obj, exc_obj, exc_obj, exc_obj,
                                         exc_obj, exc_obj, exc_obj, exc_obj,
                                         exc_obj, exc_obj, exc_obj, exc_obj,
                                         exc_obj, exc_obj, exc_obj, exc_obj,
                                         exc_obj, exc_obj, exc_obj, exc_obj,
                                         {}]
                handler.create_or_update_rule(
                    'security_group_rule.create.end', payload, None)
                # Verified we bailed out after 20 retires.
                self.assertEqual(20, mock_get.call_count)
                self.assertEqual(20, mock_save.call_count)

    def test_rule_delete_exception(self):
        # Set up the return documents.
        payload = {'security_group_rule_id': ID1}
        doc_get = {'_source': {'security_group_rules': [], 'id': 1},
                   '_version': 1}
        doc_nest = {'hits': {'hits': [{
                    '_id': 123456789,
                    '_source': {'security_group_rules': []},
                    '_version': 1}]}}

        handler = self.plugin.get_notification_handler()
        with mock.patch.object(self.plugin.index_helper,
                               'get_docs_by_nested_field') as mo_nest:
            with mock.patch.object(self.plugin.index_helper,
                                   'get_document') as mock_get:
                with mock.patch.object(self.plugin.index_helper,
                                       'save_document') as mock_save:
                    mo_nest.return_value = doc_nest
                    mock_get.return_value = doc_get
                    exc_obj = helpers.BulkIndexError(
                        "Version conflict", [{'index': {
                            "_id": "1", "error": "Some error", "status": 409}}]
                    )

                    # 1 retry (exception).
                    mock_save.side_effect = [exc_obj, {}]
                    handler.delete_rule(
                        'security_group_rule.delete.end', payload, None)
                    # 1 retry +  1 success = 2 calls.
                    self.assertEqual(1, mo_nest.call_count)
                    self.assertEqual(1, mock_get.call_count)
                    self.assertEqual(2, mock_save.call_count)

                    # 24 retries (exceptions) that exceed the retry limit.
                    # Not all retries will be used.
                    mo_nest.reset_mock()
                    mock_get.reset_mock()
                    mock_save.reset_mock()
                    mock_save.side_effect = [exc_obj, exc_obj, exc_obj,
                                             exc_obj, exc_obj, exc_obj,
                                             exc_obj, exc_obj, exc_obj,
                                             exc_obj, exc_obj, exc_obj,
                                             exc_obj, exc_obj, exc_obj,
                                             exc_obj, exc_obj, exc_obj,
                                             exc_obj, exc_obj, exc_obj,
                                             exc_obj, exc_obj, exc_obj,
                                             {}]
                    handler.delete_rule(
                        'security_group_rule.delete.end', payload, None)
                    # Verified we bailed out after 20 retires.
                    self.assertEqual(1, mo_nest.call_count)
                    self.assertEqual(20, mock_get.call_count)
                    self.assertEqual(20, mock_save.call_count)

    @mock.patch('searchlight.elasticsearch.plugins.utils.get_now_str')
    def test_serialize(self, mock_now):
        """Serializing the Security Group data will result in two new fields
           added to the record: project_id (copied from tenant_id) and
           updated_id (set to now_str). Verify this is done correctly.
        """
        mock_now.return_value = _now_str

        serialized = self.plugin.serialize(self.security_group)
        self.assertEqual(TENANT1, serialized['project_id'])
        self.assertEqual(_now_str, serialized['updated_at'])



# Copyright (c) 2015-2016 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock

from searchlight.listener import NotificationEndpoint
from searchlight.pipeline import PipelineManager
from searchlight.tests import functional
from searchlight.tests.functional import test_listener
from searchlight.tests import utils


PROJECT1 = "34518c16d95e40a19b1a95c1916d8335"
PROJECT2 = "78d384ce822d420084d706a167f84c95"

EVENT_TENANT1 = "80264096ac454d3d904002491fafe2ec"
EVENT_RS_ID = "cfe39618-49da-4877-914b-f25ef0fb3dc1"
EVENT_ZONE_ID = "3081593e-10ca-408c-af77-1397e689c177"


class TestDesignatePlugins(functional.FunctionalTest):
    def setUp(self):
        super(TestDesignatePlugins, self).setUp()
        self.zone_plugin = self.initialized_plugins['OS::Designate::Zone']
        self.rs_plugin = self.initialized_plugins['OS::Designate::RecordSet']
        self.zone_objects = self._load_fixture_data('load/zones.json')
        self.rs_objects = self._load_fixture_data('load/recordsets.json')

    def test_zone_rbac(self):
        self._index(self.zone_plugin, self.zone_objects)
        response, json_content = self._search_request(
            {"query": {"match_all": {}}, "type": "OS::Designate::Zone"},
            PROJECT1)

        self.assertEqual(1, len(json_content['hits']['hits']))
        # Shouldn't see the admin-example zone belonging to project2
        self.assertEqual('example.com.',
                         json_content['hits']['hits'][0]['_source']['name'])

    def test_recordset_rbac(self):
        self._index(self.rs_plugin, self.rs_objects)
        response, json_content = self._search_request(
            {"query": {"match_all": {}}, "type": "OS::Designate::RecordSet"},
            PROJECT1)
        self.assertEqual(2, len(json_content['hits']['hits']))

        hit_zone_names = set(h['_source']['zone_name']
                             for h in json_content['hits']['hits'])
        # Shouldn't see any admin-example results
        self.assertEqual(set(['example.com.']), hit_zone_names)


class TestDesignateListener(test_listener.TestSearchListenerBase):
    def __init__(self, *args, **kwargs):
        super(TestDesignateListener, self).__init__(*args, **kwargs)
        self.zone_events = self._load_fixture_data('events/zones.json')
        self.recordset_events = self._load_fixture_data(
            'events/recordsets.json')
        self.record_events = self._load_fixture_data('events/records.json')

    def setUp(self):
        super(TestDesignateListener, self).setUp()

        self.zones_plugin = self.initialized_plugins['OS::Designate::Zone']
        self.recordsets_plugin = self.initialized_plugins[
            'OS::Designate::RecordSet']

        '''openstack_client_mod = "searchlight.elasticsearch.plugins." \
                               "openstack_clients.get_designateclient"
        osclient_patcher = mock.patch(
            openstack_client_mod,
            mock_designate_pyclient.get_fake_designate_client
        )
        osclient_patcher.start()
        self.addCleanup(osclient_patcher.stop)'''

        notification_plugins = {
            plugin.document_type: utils.StevedoreMock(plugin)
            for plugin in (self.zones_plugin, self.recordsets_plugin)}
        self.notification_endpoint = NotificationEndpoint(
            notification_plugins,
            PipelineManager(notification_plugins)
        )

        self.listener_alias = self.zones_plugin.alias_name_listener

    @mock.patch('searchlight.elasticsearch.plugins.designate._get_recordsets')
    def test_zone_cud_events(self, mock_get_recordsets):
        """Test CUD operations for zones"""
        mock_get_recordsets.return_value = []

        create_event = self.zone_events["dns.zone.create"]
        self._send_event_to_listener(create_event, self.listener_alias)
        result = self._verify_event_processing(
            create_event)
        verification_keys = ['id', 'name']
        self._verify_result(create_event, verification_keys, result)

        mock_get_recordsets.assert_called_with(EVENT_ZONE_ID)

        # Update zone
        update_event = self.zone_events['dns.zone.update']
        self._send_event_to_listener(update_event, self.listener_alias)
        result = self._verify_event_processing(
            update_event)
        verification_keys = ['ttl']
        self._verify_result(update_event, verification_keys, result)

        # Delete Zone
        delete_event = self.zone_events['dns.zone.delete']
        self._send_event_to_listener(delete_event, self.listener_alias)
        self._verify_event_processing(
            delete_event, count=0)

    @mock.patch('searchlight.elasticsearch.plugins.designate._get_recordsets')
    def test_zone_create_recordsets(self, mock_get_recordsets):
        """When a zone's created, the event processing requests recordsets
        created as part of the process.
        """
        recordset_objects = self._load_fixture_data('load/recordsets.json')

        create_event = self.zone_events["dns.zone-recordset.create"]
        zone_id = create_event['payload']['id']

        # Simulate what would come back from the recordset API call
        zone_records = list(filter(
            lambda r: r['zone_id'] == zone_id,
            recordset_objects))
        self.assertEqual(2, len(zone_records))
        mock_get_recordsets.return_value = zone_records

        self._send_event_to_listener(create_event, self.listener_alias)
        mock_get_recordsets.assert_called_with(zone_id)

        # Expect 3 results; the zone and the recordsets
        self._verify_event_processing(create_event, count=3)

        query = {"filter": {"term": {"zone_id": zone_id}},
                 "type": self.recordsets_plugin.get_document_type()}
        response, search_results = self._search_request(query, EVENT_TENANT1)
        self.assertEqual(2, search_results['hits']['total'])

    def test_recordset_cud_events(self):
        create_event = self.recordset_events["dns.recordset.create"]
        self._send_event_to_listener(create_event, self.listener_alias)
        result = self._verify_event_processing(
            create_event)
        verification_keys = ['id', 'name', 'zone_id', 'type']
        self._verify_result(create_event, verification_keys, result)

        query = {"filter": {"term": {"records": "10.0.0.1"}},
                 "type": self.recordsets_plugin.get_document_type()}
        response, search_results = self._search_request(query, EVENT_TENANT1)
        self.assertEqual(1, search_results['hits']['total'])
        self.assertEqual(EVENT_RS_ID,
                         search_results['hits']['hits'][0]['_source']['id'])

        query['filter']['term']['records'] = "10.0.0.2"
        response, search_results = self._search_request(query, EVENT_TENANT1)
        self.assertEqual(1, search_results['hits']['total'])
        self.assertEqual(EVENT_RS_ID,
                         search_results['hits']['hits'][0]['_source']['id'])

        # Update zone
        update_event = self.recordset_events['dns.recordset.update']
        self._send_event_to_listener(update_event, self.listener_alias)
        result = self._verify_event_processing(
            update_event)
        verification_keys = ['version']
        self._verify_result(update_event, verification_keys, result)

        delete_event = self.recordset_events['dns.recordset.delete']
        self._send_event_to_listener(delete_event, self.listener_alias)
        self._verify_event_processing(
            delete_event, count=0)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import mock

import novaclient.v2.flavors as novaclient_flavors

from searchlight.elasticsearch.plugins.nova import\
    flavors as flavors_plugin
import searchlight.tests.utils as test_utils

_now = datetime.datetime.utcnow()
updated_now = _now.strftime('%Y-%m-%dT%H:%M:%SZ')

fake_version_list = [test_utils.FakeVersion('2.1'),
                     test_utils.FakeVersion('2.1')]

nova_version_getter = 'novaclient.v2.client.versions.VersionManager.list'


def _flavor_fixture(name, id, ram, vcpus, swap, disk, is_public, ephemeral,
                    extra_specs=None):
    attrs = {
        "name": name,
        "id": id,
        "links": [
            {'href': 'http://127.0.0.1:8774/v2.1/flavors/test', 'rel': 'self'},
            {'href': 'http://127.0.0.1:8774/flavors/test', 'rel': 'bookmark'}
        ],
        "ram": ram,
        "disk": disk,
        "vcpus": vcpus,
        "swap": swap,
        "OS-FLV-EXT-DATA:ephemeral": ephemeral,
        "os-flavor-access:is_public": is_public,
        "OS-FLV-DISABLED:disabled": False,
        "rxtx_factor": 1.0,
    }

    flavor = mock.Mock(spec=novaclient_flavors.Flavor, **attrs)
    flavor.to_dict.return_value = attrs
    flavor.get_keys.return_value = extra_specs or {}
    return flavor


class TestFlavorLoaderPlugin(test_utils.BaseTestCase):
    def setUp(self):
        super(TestFlavorLoaderPlugin, self).setUp()
        self.plugin = flavors_plugin.FlavorIndex()
        self._create_fixtures()

    def _create_fixtures(self):
        self.flavor1 = _flavor_fixture("test_flavor",
                                       "100", 512, 1, 0, 10, True, 0,
                                       extra_specs={"key1": "value1"})
        self.versioned_flavor = {
            'disabled': False,
            'root_gb': 10,
            'name': 'test_flavor',
            'ephemeral_gb': 0,
            'memory_mb': 512,
            'vcpus': 1,
            'swap': 0,
            'rxtx_factor': 1.0,
            'is_public': True,
            'flavorid': '100',
            'extra_specs': {'key1': 'value1'},
            'vcpu_weight': 0
        }

    def test_index_name(self):
        self.assertEqual('searchlight', self.plugin.resource_group_name)

    def test_document_type(self):
        self.assertEqual('OS::Nova::Flavor',
                         self.plugin.get_document_type())

    @mock.patch(nova_version_getter, return_value=fake_version_list)
    def test_serialize(self, mock_version):
        expected = {
            'OS-FLV-DISABLED:disabled': False,
            'OS-FLV-EXT-DATA:ephemeral': 0,
            'disk': 10,
            'extra_specs': {'key1': 'value1'},
            'id': '100',
            'name': 'test_flavor',
            'os-flavor-access:is_public': True,
            'ram': 512,
            'rxtx_factor': 1.0,
            'swap': 0,
            'tenant_access': None,
            'updated_at': updated_now,
            'vcpus': 1
        }
        with mock.patch('searchlight.elasticsearch.plugins.utils.get_now_str',
                        return_value=updated_now):
            serialized1 = self.plugin.serialize(self.flavor1)
            serialized2 = self.plugin.serialize(self.versioned_flavor)
        self.assertEqual(expected, serialized1)
        self.assertEqual(expected, serialized2)

    def test_notification_events(self):
        handler = self.plugin.get_notification_handler()
        self.assertEqual(
            set(['flavor.create',
                 'flavor.update',
                 'flavor.delete']),
            set(handler.get_event_handlers().keys())
        )



# Copyright (c) 2016 Hewlett-Packard Enterprise Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from searchlight.common import resource_types
from searchlight.elasticsearch.plugins import base
from searchlight.elasticsearch.plugins.neutron import add_rbac
from searchlight.elasticsearch.plugins.neutron import notification_handlers
from searchlight.elasticsearch.plugins.neutron import serialize_network
from searchlight.elasticsearch.plugins import openstack_clients


class NetworkIndex(base.IndexBase):
    NotificationHandlerCls = notification_handlers.NetworkHandler

    ADMIN_ONLY_FIELDS = ['provider:*']

    def __init__(self):
        super(NetworkIndex, self).__init__()

    @classmethod
    def get_document_type(self):
        return resource_types.NEUTRON_NETWORK

    def get_mapping(self):
        return {
            'dynamic': False,
            'properties': {
                # Availability zones are not present because they're not
                # contained in notifications
                'admin_state_up': {'type': 'boolean'},
                'availability_zone_hints': {'type': 'string',
                                            'index': 'not_analyzed'},
                'availability_zones': {'type': 'string',
                                       'index': 'not_analyzed'},
                'created_at': {'type': 'date'},
                'description': {'type': 'string'},
                'id': {'type': 'string', 'index': 'not_analyzed'},
                'ipv4_address_scope': {'type': 'string',
                                       'index': 'not_analyzed'},
                'ipv6_address_scope': {'type': 'string',
                                       'index': 'not_analyzed'},
                'mtu': {'type': 'integer'},
                'name': {
                    'type': 'string',
                    'fields': {
                        'raw': {'type': 'string', 'index': 'not_analyzed'}
                    }
                },
                'port_security_enabled': {'type': 'boolean'},
                'project_id': {'type': 'string', 'index': 'not_analyzed'},
                'provider:network_type': {'type': 'string'},
                'provider:physical_network': {'type': 'string'},
                'provider:segmentation_id': {'type': 'integer'},
                'router:external': {'type': 'boolean'},
                'shared': {'type': 'boolean'},
                'status': {'type': 'string', 'index': 'not_analyzed'},
                'tenant_id': {'type': 'string', 'index': 'not_analyzed'},
                'members': {'type': 'string', 'index': 'not_analyzed'},
                'rbac_policy': {
                    'type': 'nested',
                    'properties': {
                        'rbac_id': {'type': 'string', 'index': 'not_analyzed'},
                        'tenant': {'type': 'string', 'index': 'not_analyzed'}
                    }
                },
                'updated_at': {'type': 'date'}
            },
            "_meta": {
                "project_id": {
                    "resource_type": resource_types.KEYSTONE_PROJECT
                },
                "tenant_id": {
                    "resource_type": resource_types.KEYSTONE_PROJECT
                }
            }
        }

    @property
    def admin_only_fields(self):
        from_conf = super(NetworkIndex, self).admin_only_fields
        return from_conf + NetworkIndex.ADMIN_ONLY_FIELDS

    @property
    def facets_with_options(self):
        return ('provider:network_type', 'provider:physical_network', 'status',
                'port_security_enabled', 'shared', 'availability_zones')

    @property
    def facets_excluded(self):
        """A map of {name: allow_admin} that indicate which
        fields should not be offered as facet options, or those that should
        only be available to administrators.
        """
        return {'tenant_id': True, 'project_id': True}

    @property
    def resource_allowed_policy_target(self):
        # Neutron only supports policy for individual networks
        return None

    @property
    def service_type(self):
        return 'network'

    def _get_rbac_field_filters(self, request_context):
        """Return any RBAC field filters to be injected into an indices
        query. Document type will be added to this list.
        """
        return [
            {'term': {'tenant_id': request_context.owner}},
            {'terms': {'members': [request_context.owner, '*']}},
            {'term': {'router:external': True}},
            {'term': {'shared': True}}
        ]

    def get_rbac_policies(self):
        policies = defaultdict(list)
        policy_list = self.get_rbac_objects()
        for policy in policy_list:
            policies[policy['object_id']].append(policy)
        return policies

    def get_objects(self):
        """Generator that lists all networks owned by all tenants."""
        # Neutronclient handles pagination itself; list_networks is a generator
        policies = self.get_rbac_policies()
        neutron_client = openstack_clients.get_neutronclient()
        for network in neutron_client.list_networks()['networks']:
            network['members'] = []
            network['rbac_policy'] = []
            for policy in policies[network['id']]:
                add_rbac(network, policy['target_tenant'], policy['id'])
            yield network

    def serialize(self, network):
        return serialize_network(network)

    def filter_result(self, hit, request_context):
        # The mapping contains internal fields related to RBAC policy.
        # Remove them.
        source = hit['_source']
        source.pop('rbac_policy', None)
        source.pop('members', None)

    def get_rbac_objects(self):
        """Generator that lists all RBAC policies for all tenants."""
        valid_actions = notification_handlers.RBAC_VALID_ACTIONS
        neutron_client = openstack_clients.get_neutronclient()
        policies = neutron_client.list_rbac_policies()['rbac_policies']
        for policy in [p for p in policies if
                       p['object_type'] == 'network' and
                       p['action'] in valid_actions]:
            yield policy



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import glanceclient.exc
from oslo_log import log as logging

from searchlight.elasticsearch.plugins import base
from searchlight.elasticsearch.plugins.glance \
    import serialize_glance_image_members
from searchlight.elasticsearch.plugins.glance \
    import serialize_glance_notification
from searchlight import pipeline

LOG = logging.getLogger(__name__)


class ImageHandler(base.NotificationBase):

    def __init__(self, *args, **kwargs):
        super(ImageHandler, self).__init__(*args, **kwargs)
        self.image_delete_keys = ['deleted_at', 'deleted',
                                  'is_public', 'properties']

    @classmethod
    def _get_notification_exchanges(cls):
        return ['glance']

    def get_event_handlers(self):
        return {
            "image.create": self.create_or_update,
            "image.update": self.create_or_update,
            "image.delete": self.delete,
            "image.member.create": self.sync_members,
            "image.member.update": self.sync_members,
            "image.member.delete": self.sync_members
        }

    def get_log_fields(self, event_type, payload):
        return (
            ('id', payload.get('id')),
            ('owner', payload.get('owner'))
        )

    def serialize_notification(self, notification):
        return serialize_glance_notification(notification)

    def create_or_update(self, event_type, payload, timestamp):
        image_id = payload['id']
        try:
            image_payload = self.serialize_notification(payload)
            self.index_helper.save_document(
                image_payload,
                version=self.get_version(image_payload, timestamp))
            return pipeline.IndexItem(self.index_helper.plugin,
                                      event_type,
                                      payload,
                                      image_payload)
        except glanceclient.exc.NotFound:
            LOG.warning("Image %s not found; deleting" % image_id)
            return self.delete(event_type, payload, timestamp)

    def delete(self, event_type, payload, timestamp):
        image_id = payload['id']
        try:
            version = self.get_version(payload, timestamp,
                                       preferred_date_field='deleted_at')
            self.index_helper.delete_document(
                {'_id': image_id, '_version': version})
            return pipeline.DeleteItem(self.index_helper.plugin,
                                       event_type,
                                       payload,
                                       image_id)
        except Exception as exc:
            LOG.error(
                'Error deleting image %(image_id)s from index: %(exc)s' %
                {'image_id': image_id, 'exc': exc})

    def sync_members(self, event_type, payload, timestamp):
        image_id = payload['image_id']
        image_es = self.index_helper.get_document(image_id,
                                                  for_admin=True)

        image_payload = serialize_glance_image_members(image_es['_source'],
                                                       payload)

        self.index_helper.save_document(image_payload)
        return pipeline.IndexItem(self.index_helper.plugin,
                                  event_type,
                                  payload,
                                  image_payload)



# Copyright 2016 Intel Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from searchlight.elasticsearch.plugins.utils import normalize_es_document
import stevedore

LOG = logging.getLogger(__name__)


class Pipeline(object):

    def __init__(self, plugin, publisher):
        self.plugin = plugin
        self.publisher = publisher
        self.notification_handler = self.plugin.get_notification_handler()

    def supported_events(self):
        if self.notification_handler:
            return self.notification_handler.\
                get_notification_supported_events()
        return []


class PipelineManager(object):

    def __init__(self, plugins):
        self.pipelines = []
        for plugin in plugins.values():
            publishers = []
            if plugin.obj.publishers:
                for publisher_name in plugin.obj.publishers:
                    try:
                        publishers.append(get_publisher(publisher_name))
                    except Exception as exc:
                        LOG.warning(
                            'Unable to load publisher %(publisher)s '
                            'of plugin %(plugin)s: %(exc)s' % {
                                'publisher': publisher_name,
                                'plugin': plugin.obj.name,
                                'exc': exc}
                        )

            for publisher in publishers:
                pipeline = Pipeline(plugin.obj, publisher)
                self.pipelines.append(pipeline)

    def publish(self, items):
        for pipeline in self.pipelines:
            for item in items:
                if item.event_type in pipeline.supported_events():
                    pipeline.publisher.publish(item)


class PipelineItem(object):
    """
    Base class to store notification and resource info
    """

    def __init__(self, plugin, event_type, payload):
        """
        :param plugin: Plugin instance of modified resource.
        :param event_type: event type of the notification.
        :param payload: notification payload received. Note that payload may
        contain sensitive information shouldn't be passed on.
        """
        self.plugin = plugin
        self.resource_type = plugin.get_document_type()
        self.payload = payload
        self.event_type = event_type


class IndexItem(PipelineItem):

    def __init__(self, plugin, event_type, payload, doc):
        """
        :param plugin: Plugin instance of modified resource.
        :param event_type: event type of the notification.
        :param payload: notification payload received. Note that payload may
        contain sensitive information shouldn't be passed on.
        :param doc: resource document passed to publishers.
        """
        super(IndexItem, self).__init__(plugin, event_type, payload)
        self.doc_id = doc[plugin.get_document_id_field()]
        self.doc = normalize_es_document(doc, plugin)


class DeleteItem(PipelineItem):

    def __init__(self, resource_type, event_type, payload, doc_id):
        """
        :param resource_type: Heat resource type of modified resources,
        see searchlight/common/resource_types.py for more detail.
        :param event_type: event type of the notification.
        :param payload: notification payload received, Note that payload may
        contain sensitive information shouldn't be passed on.
        :param doc_id: id of document to be deleted.
        """
        super(DeleteItem, self).__init__(resource_type, event_type, payload)
        self.doc_id = doc_id


def get_publisher(url):
    namespace = 'searchlight.publisher'
    loaded_driver = stevedore.driver.DriverManager(
        namespace, url, invoke_on_load=True)
    return loaded_driver.driver
